"""Unit tests for LDAPBackend."""

import kgb
import ldap
from django.conf import settings
from django.contrib.auth.models import User
from django.test import override_settings
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.backends import LDAPBackend
from reviewboard.testing import TestCase


class TestLDAPObject(object):
    def __init__(self, *args, **kwargs):
        pass

    def set_option(self, option, value):
        pass

    def start_tls_s(self):
        pass

    def simple_bind_s(self, *args, **kwargs):
        pass

    def bind_s(self, *args, **kwargs):
        pass

    def search_s(self, *args, **kwargs):
        pass


class LDAPAuthBackendTests(kgb.SpyAgency, TestCase):
    """Unit tests for the LDAP authentication backend."""

    DEFAULT_FILTER_STR = '(objectClass=*)'

    def setUp(self):
        super(LDAPAuthBackendTests, self).setUp()

        # These settings will get overridden on future test runs, since
        # they'll be reloaded from siteconfig.
        settings.LDAP_BASE_DN = 'CN=admin,DC=example,DC=com'
        settings.LDAP_GIVEN_NAME_ATTRIBUTE = 'givenName'
        settings.LDAP_SURNAME_ATTRIBUTE = 'sn'
        settings.LDAP_EMAIL_ATTRIBUTE = 'email'
        settings.LDAP_UID = 'uid'
        settings.LDAP_UID_MASK = None
        settings.LDAP_FULL_NAME_ATTRIBUTE = None

        self.backend = LDAPBackend()

        self.spy_on(ldap.initialize,
                    call_fake=lambda uri, *args, **kwargs: TestLDAPObject(uri))

    @add_fixtures(['test_users'])
    def test_authenticate_with_valid_credentials(self):
        """Testing LDAPBackend.authenticate with valid credentials"""
        self.spy_on(TestLDAPObject.bind_s,
                    owner=TestLDAPObject)
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([
                        ('CN=Doc Dwarf,OU=MyOrg,DC=example,DC=COM', {}),
                    ]))

        user = self.backend.authenticate(request=None,
                                         username='doc',
                                         password='mypass')
        self.assertIsNotNone(user)

        self.assertEqual(user.username, 'doc')
        self.assertEqual(user.first_name, 'Doc')
        self.assertEqual(user.last_name, 'Dwarf')
        self.assertEqual(user.email, 'doc@example.com')
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)

        self.assertSpyCalledWith(TestLDAPObject.bind_s,
                                 'CN=Doc Dwarf,OU=MyOrg,DC=example,DC=COM',
                                 'mypass')
        self.assertSpyCalledWith(TestLDAPObject.search_s,
                                 'CN=admin,DC=example,DC=com',
                                 ldap.SCOPE_SUBTREE,
                                 '(uid=doc)')

    def test_authenticate_with_invalid_credentials(self):
        """Testing LDAPBackend.authenticate with invalid credentials"""
        self.spy_on(TestLDAPObject.bind_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpRaise(ldap.INVALID_CREDENTIALS()))
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([
                        ('CN=Doc Dwarf,OU=MyOrg,DC=example,DC=COM', {}),
                    ]))

        user = self.backend.authenticate(request=None,
                                         username='doc',
                                         password='mypass')
        self.assertIsNone(user)

        self.assertSpyCalledWith(
            TestLDAPObject.bind_s,
            'CN=Doc Dwarf,OU=MyOrg,DC=example,DC=COM',
            'mypass')
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'CN=admin,DC=example,DC=com',
            ldap.SCOPE_SUBTREE,
            '(uid=doc)')

    def test_authenticate_with_invalid_credentials_ok_anon_bind(self):
        """Testing LDAPBackend.authenticate with invalid credentials but
        anonymous bind without error
        """
        self.spy_on(TestLDAPObject.bind_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn(None))
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpRaise(ldap.INVALID_CREDENTIALS()))

        self.create_user(username='local-user')

        user = self.backend.authenticate(request=None,
                                         username='local-user',
                                         password='mypass')
        self.assertIsNone(user)

        self.assertSpyNotCalled(TestLDAPObject.bind_s)
        self.assertSpyCalledWith(
            TestLDAPObject.search_s,
            'CN=admin,DC=example,DC=com',
            ldap.SCOPE_SUBTREE,
            '(uid=local-user)')

    def test_authenticate_with_ldap_error(self):
        """Testing LDAPBackend.authenticate with LDAP error"""
        self.spy_on(TestLDAPObject.bind_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpRaise(ldap.LDAPError()))
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([
                        ('CN=Doc Dwarf,OU=MyOrg,DC=example,DC=COM', {}),
                    ]))

        user = self.backend.authenticate(request=None,
                                         username='doc',
                                         password='mypass')
        self.assertIsNone(user)

        self.assertSpyCalledWith(TestLDAPObject.bind_s,
                                 'CN=Doc Dwarf,OU=MyOrg,DC=example,DC=COM',
                                 'mypass')
        self.assertSpyCalledWith(TestLDAPObject.search_s,
                                 'CN=admin,DC=example,DC=com',
                                 ldap.SCOPE_SUBTREE,
                                 '(uid=doc)')

    def test_authenticate_with_exception(self):
        """Testing LDAPBackend.authenticate with unexpected exception"""
        self.spy_on(TestLDAPObject.bind_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpRaise(Exception('oh no!')))
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([
                        ('CN=Doc Dwarf,OU=MyOrg,DC=example,DC=COM', {}),
                    ]))

        user = self.backend.authenticate(request=None,
                                         username='doc',
                                         password='mypass')
        self.assertIsNone(user)

        self.assertSpyCalledWith(TestLDAPObject.bind_s,
                                 'CN=Doc Dwarf,OU=MyOrg,DC=example,DC=COM',
                                 'mypass')
        self.assertSpyCalledWith(TestLDAPObject.search_s,
                                 'CN=admin,DC=example,DC=com',
                                 ldap.SCOPE_SUBTREE,
                                 '(uid=doc)')

    @add_fixtures(['test_users'])
    def test_get_or_create_user_with_existing_user(self):
        """Testing LDAPBackend.get_or_create_user with existing user"""
        original_count = User.objects.count()
        user = User.objects.get(username='doc')
        result = self.backend.get_or_create_user(username='doc', request=None)

        self.assertEqual(original_count, User.objects.count())
        self.assertEqual(user, result)

    def test_get_or_create_user_in_ldap(self):
        """Testing LDAPBackend.get_or_create_user with new user found in LDAP
        """
        user_dn = 'CN=Bob BobBob,OU=MyOrg,DC=example,DC=COM'

        self.spy_on(
            TestLDAPObject.search_s,
            owner=TestLDAPObject,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ('CN=admin,DC=example,DC=com',
                             ldap.SCOPE_SUBTREE, '(uid=doc)'),
                    'call_fake': lambda *args, **kwargs: [(user_dn, {})],
                },
                {
                    'args': (user_dn, ldap.SCOPE_BASE),
                    'call_fake': lambda *args, **kwargs: [
                        (user_dn, {
                            'givenName': [b'Bob'],
                            'sn': [b'BobBob'],
                            'email': [b'imbob@example.com'],
                        }),
                    ],
                },
            ]))

        self.assertEqual(User.objects.count(), 0)

        user = self.backend.get_or_create_user(username='doc', request=None)
        self.assertIsNotNone(user)
        self.assertEqual(User.objects.count(), 1)

        self.assertEqual(user.username, 'doc')
        self.assertEqual(user.first_name, 'Bob')
        self.assertEqual(user.last_name, 'BobBob')
        self.assertEqual(user.email, 'imbob@example.com')
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)

    def test_get_or_create_user_not_in_ldap(self):
        """Testing LDAPBackend.get_or_create_user with new user not found in
        LDAP
        """
        self.spy_on(TestLDAPObject.search_s,
                    owner=TestLDAPObject,
                    op=kgb.SpyOpReturn([]))

        self.assertEqual(User.objects.count(), 0)

        user = self.backend.get_or_create_user(username='doc', request=None)

        self.assertIsNone(user)
        self.assertEqual(User.objects.count(), 0)

        self.assertSpyCalledWith(TestLDAPObject.search_s,
                                 'CN=admin,DC=example,DC=com',
                                 ldap.SCOPE_SUBTREE,
                                 '(uid=doc)')

    @override_settings(LDAP_GIVEN_NAME_ATTRIBUTE='myFirstName')
    def test_get_or_create_user_with_empty_info(self):
        """Testing LDAPBackend.get_or_create_user with empty information in
        LDAP
        """
        user_dn = 'CN=Bob,OU=MyOrg,DC=example,DC=COM'

        self.spy_on(
            TestLDAPObject.search_s,
            owner=TestLDAPObject,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ('CN=admin,DC=example,DC=com',
                             ldap.SCOPE_SUBTREE, '(uid=doc)'),
                    'call_fake': lambda *args, **kwargs: [(user_dn, {})],
                },
                {
                    'args': (user_dn, ldap.SCOPE_BASE),
                    'call_fake': lambda *args, **kwargs: [(user_dn, {})]
                },
            ]))

        self.assertEqual(User.objects.count(), 0)

        user = self.backend.get_or_create_user(username='doc',
                                               request=None)

        self.assertIsNotNone(user)
        self.assertEqual(User.objects.count(), 1)

        self.assertEqual(user.first_name, 'doc')
        self.assertEqual(user.last_name, '')
        self.assertEqual(user.email, '')

    @override_settings(LDAP_GIVEN_NAME_ATTRIBUTE='myFirstName')
    def test_get_or_create_user_with_given_name_attr(self):
        """Testing LDAPBackend.get_or_create_user with
        LDAP_GIVEN_NAME_ATTRIBUTE
        """
        user_dn = 'CN=Bob,OU=MyOrg,DC=example,DC=COM'

        self.spy_on(
            TestLDAPObject.search_s,
            owner=TestLDAPObject,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ('CN=admin,DC=example,DC=com',
                             ldap.SCOPE_SUBTREE, '(uid=doc)'),
                    'call_fake': lambda *args, **kwargs: [(user_dn, {})],
                },
                {
                    'args': (user_dn, ldap.SCOPE_BASE),
                    'call_fake': lambda *args, **kwargs: [
                        (user_dn, {
                            'myFirstName': [b'Bob'],
                            'email': [b'imbob@example.com'],
                        }),
                    ],
                },
            ]))

        self.assertEqual(User.objects.count(), 0)

        user = self.backend.get_or_create_user(username='doc',
                                               request=None)

        self.assertIsNotNone(user)
        self.assertEqual(User.objects.count(), 1)

        self.assertEqual(user.first_name, 'Bob')
        self.assertEqual(user.last_name, '')
        self.assertEqual(user.email, 'imbob@example.com')

    @override_settings(LDAP_SURNAME_ATTRIBUTE='myLastName')
    def test_get_or_create_user_with_surname_attr(self):
        """Testing LDAPBackend.get_or_create_user with LDAP_SURNAME_ATTRIBUTE
        """
        user_dn = 'CN=Bob,OU=MyOrg,DC=example,DC=COM'

        self.spy_on(
            TestLDAPObject.search_s,
            owner=TestLDAPObject,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ('CN=admin,DC=example,DC=com',
                             ldap.SCOPE_SUBTREE, '(uid=doc)'),
                    'call_fake': lambda *args, **kwargs: [(user_dn, {})],
                },
                {
                    'args': (user_dn, ldap.SCOPE_BASE),
                    'call_fake': lambda *args, **kwargs: [
                        (user_dn, {
                            'givenName': [b'Bob'],
                            'myLastName': [b'Bub'],
                            'email': [b'imbob@example.com'],
                        }),
                    ],
                },
            ]))

        self.assertEqual(User.objects.count(), 0)

        user = self.backend.get_or_create_user(username='doc',
                                               request=None)

        self.assertIsNotNone(user)
        self.assertEqual(User.objects.count(), 1)

        self.assertEqual(user.first_name, 'Bob')
        self.assertEqual(user.last_name, 'Bub')
        self.assertEqual(user.email, 'imbob@example.com')

    @override_settings(LDAP_FULL_NAME_ATTRIBUTE='fn')
    def test_get_or_create_user_with_fullname(self):
        """Testing LDAPBackend.get_or_create_user with LDAP_FULL_NAME_ATTRIBUTE
        """
        user_dn = 'CN=Bob,OU=MyOrg,DC=example,DC=COM'

        self.spy_on(
            TestLDAPObject.search_s,
            owner=TestLDAPObject,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ('CN=admin,DC=example,DC=com',
                             ldap.SCOPE_SUBTREE, '(uid=doc)'),
                    'call_fake': lambda *args, **kwargs: [(user_dn, {})],
                },
                {
                    'args': (user_dn, ldap.SCOPE_BASE),
                    'call_fake': lambda *args, **kwargs: [
                        (user_dn, {
                            'fn': [b'Bob Bab Bub'],
                            'email': [b'imbob@example.com'],
                        }),
                    ],
                },
            ]))

        self.assertEqual(User.objects.count(), 0)

        user = self.backend.get_or_create_user(username='doc',
                                               request=None)

        self.assertIsNotNone(user)
        self.assertEqual(User.objects.count(), 1)

        self.assertEqual(user.first_name, 'Bob')
        self.assertEqual(user.last_name, 'Bab Bub')
        self.assertEqual(user.email, 'imbob@example.com')

    @override_settings(LDAP_FULL_NAME_ATTRIBUTE='fn')
    def test_get_or_create_user_with_fullname_without_space(self):
        """Testing LDAPBackend.get_or_create_user with LDAP_FULL_NAME_ATTRIBUTE
        and user whose full name does not contain a space
        """
        user_dn = 'CN=Bob,OU=MyOrg,DC=example,DC=COM'

        self.spy_on(
            TestLDAPObject.search_s,
            owner=TestLDAPObject,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ('CN=admin,DC=example,DC=com',
                             ldap.SCOPE_SUBTREE, '(uid=doc)'),
                    'call_fake': lambda *args, **kwargs: [(user_dn, {})],
                },
                {
                    'args': (user_dn, ldap.SCOPE_BASE),
                    'call_fake': lambda *args, **kwargs: [
                        (user_dn, {
                            'fn': [b'Bob'],
                            'email': [b'imbob@example.com'],
                        }),
                    ],
                },
            ]))

        self.assertEqual(User.objects.count(), 0)

        user = self.backend.get_or_create_user(username='doc',
                                               request=None)

        self.assertIsNotNone(user)
        self.assertEqual(User.objects.count(), 1)

        self.assertEqual(user.first_name, 'Bob')
        self.assertEqual(user.last_name, '')
        self.assertEqual(user.email, 'imbob@example.com')



"""A Review UI for diffs.

Version Added:
    7.0
"""

from __future__ import annotations

from typing import Optional, Sequence, TYPE_CHECKING

from reviewboard.diffviewer.models import FileDiff
from reviewboard.reviews.models import Comment
from reviewboard.reviews.ui.base import (ReviewUI,
                                         SerializedComment,
                                         SerializedCommentBlocks)

if TYPE_CHECKING:
    from django.http import HttpRequest

    from reviewboard.reviews.models import ReviewRequest


class SerializedDiffComment(SerializedComment):
    """Serialized comment data for diffs.

    This must be kept in sync with the definitions in
    :file:`reviewboard/static/rb/js/reviews/models/commentData.ts`.

    Version Added:
        7.0
    """

    #: The line number that the comment starts on.
    line: int

    #: The number of lines that the comment spans.
    num_lines: int


class DiffReviewUI(ReviewUI[FileDiff, Comment, SerializedDiffComment]):
    """A Review UI for diffs.

    Version Added:
        7.0
    """

    name = 'Diff'

    js_model_class: str = 'RB.DiffReviewable'
    js_view_class: str = 'RB.DiffReviewableView'

    ######################
    # Instance variables #
    ######################

    #: The FileDiff for the base change in a commit range.
    base_filediff: Optional[FileDiff]

    #: The tip FileDiff when viewing an interdiff.
    interfilediff: Optional[FileDiff]

    def __init__(
        self,
        *,
        review_request: ReviewRequest,
        obj: FileDiff,
        base_filediff: Optional[FileDiff],
        interfilediff: Optional[FileDiff],
        request: HttpRequest,
    ) -> None:
        """Initialize the Review UI.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request containing the object to review.

            obj (reviewboard.diffviewer.models.FileDiff):
                The object being reviewed.

            base_filediff (reviewboard.diffviewer.models.FileDiff):
                The base filediff to use, when viewing a commit range.

            interfilediff (reviewboard.diffviewer.models.FileDiff):
                The interdiff to use, when viewing an interdiff.

            request (django.http.HttpRequest):
                The HTTP request.
        """
        assert isinstance(obj, FileDiff)

        super().__init__(
            review_request=review_request,
            obj=obj)

        self.base_filediff = base_filediff
        self.interfilediff = interfilediff
        self.request = request

    def serialize_comments(
        self,
        comments: Sequence[Comment],
    ) -> SerializedCommentBlocks[SerializedDiffComment]:
        """Serialize the comments for the diff.

        Args:
            comments (list of reviewboard.reviews.models.Comment):
                The list of objects to serialize.

        Returns:
            SerializedCommentBlocks:
            The serialized comments.
        """
        result: SerializedCommentBlocks[SerializedDiffComment] = {}

        for comment in self.flat_serialized_comments(comments):
            key = f'{comment["line"]}-{comment["num_lines"]}'

            result.setdefault(key, []).append(comment)

        return result

    def serialize_comment(
        self,
        comment: Comment,
    ) -> SerializedDiffComment:
        """Serialize a comment.

        Args:
            comment (reviewboard.reviews.models.Comment):
                The comment to serialize.

        Returns:
            SerializedDiffComment:
            The serialized comment.
        """
        return {
            **super().serialize_comment(comment),
            'line': comment.first_line,
            'num_lines': comment.num_lines,
        }



"""Delete file_count fields from DiffCommit and DiffSet.

Version Added:
    4.0
"""

from django_evolution.mutations import DeleteField


MUTATIONS = [
    DeleteField('DiffSet', 'file_count'),
    DeleteField('DiffCommit', 'file_count'),
]



"""Administration form for authentication settings."""

import logging

from django import forms
from django.utils.translation import gettext_lazy as _
from djblets.forms.widgets import AmountSelectorWidget
from djblets.siteconfig.forms import SiteSettingsForm

from reviewboard.accounts.forms.auth import LegacyAuthModuleSettingsForm
from reviewboard.admin.siteconfig import load_site_config


logger = logging.getLogger(__name__)


class AuthenticationSettingsForm(SiteSettingsForm):
    """Authentication settings for Review Board.

    Attributes:
        auth_backend_forms (dict):
            A mapping of authentication backend IDs to settings form
            instances.
    """

    CUSTOM_AUTH_ID = 'custom'
    CUSTOM_AUTH_CHOICE = (CUSTOM_AUTH_ID, _('Legacy Authentication Module'))

    auth_anonymous_access = forms.BooleanField(
        label=_('Allow anonymous read-only access'),
        help_text=_('If checked, users will be able to view review requests '
                    'and diffs without logging in.'),
        required=False)

    auth_backend = forms.ChoiceField(
        label=_('Authentication Method'),
        choices=(),
        help_text=_('The method Review Board should use for authenticating '
                    'users.'),
        required=True,
        widget=forms.Select(attrs={
            'data-subform-group': 'auth-backend',
        }))

    automatic_api_token_expiration = forms.IntegerField(
        label=_('Automatic token expiration'),
        help_text=_('API tokens are automatically created when authenticating '
                    'services to Review Board via the login page. This sets '
                    'how long it will take for tokens to expire after '
                    'creation.'),
        required=False,
        widget=AmountSelectorWidget(
            unit_choices=[
                (1, _('days')),
                (7, _('weeks')),
                (30, _('months')),
                (365, _('years')),
                (None, _('Never')),
            ],
            number_attrs={
                'min': 0,
            }))

    def __init__(self, siteconfig, *args, **kwargs):
        """Initialize the settings form.

        This will load the list of available authentication backends and
        their settings forms, allowing the browser to show the appropriate
        settings form based on the selected backend.

        Args:
            siteconfig (djblets.siteconfig.models.SiteConfiguration):
                The site configuration handling the server's settings.

            *args (tuple):
                Additional positional arguments for the parent class.

            **kwargs (dict):
                Additional keyword arguments for the parent class.
        """
        super(AuthenticationSettingsForm, self).__init__(siteconfig,
                                                         *args, **kwargs)

        from reviewboard.accounts.backends import auth_backends

        self.auth_backend_forms = {}

        cur_auth_backend = (self['auth_backend'].data or
                            self.fields['auth_backend'].initial)

        if cur_auth_backend == self.CUSTOM_AUTH_ID:
            custom_auth_form = LegacyAuthModuleSettingsForm(siteconfig,
                                                            *args, **kwargs)
        else:
            custom_auth_form = LegacyAuthModuleSettingsForm(siteconfig)

        self.auth_backend_forms[self.CUSTOM_AUTH_ID] = custom_auth_form

        backend_choices = []
        builtin_auth_choice = None

        for backend in auth_backends:
            backend_id = backend.backend_id

            try:
                if backend.settings_form:
                    if cur_auth_backend == backend_id:
                        backend_form = backend.settings_form(siteconfig,
                                                             *args, **kwargs)
                    else:
                        backend_form = backend.settings_form(siteconfig)

                    self.auth_backend_forms[backend_id] = backend_form
                    backend_form.load()

                choice = (backend_id, backend.name)

                if backend_id == 'builtin':
                    builtin_auth_choice = choice
                else:
                    backend_choices.append(choice)
            except Exception as e:
                logger.exception('Error loading authentication backend %s: %s',
                                 backend_id, e)

        backend_choices.sort(key=lambda x: x[1])
        backend_choices.insert(0, builtin_auth_choice)
        backend_choices.append(self.CUSTOM_AUTH_CHOICE)
        self.fields['auth_backend'].choices = backend_choices

        from reviewboard.accounts.sso.backends import sso_backends

        self.sso_backend_forms = {}

        form_fields = self.Meta.fieldsets[0]['fields']

        for backend in sso_backends:
            backend_id = backend.backend_id

            if backend.settings_form:
                field_id = '%s_enabled' % backend_id

                try:
                    available, reason = backend.is_available()
                except Exception as e:
                    available = False
                    reason = str(e)

                self.fields[field_id] = forms.BooleanField(
                    label=_('Enable %s Authentication') % backend.name,
                    disabled=not available,
                    help_text=reason or '',
                    required=False)

                if field_id not in form_fields:
                    form_fields.append(field_id)

                form = backend.settings_form(siteconfig, *args, **kwargs)
                form.load()
                self.sso_backend_forms[backend_id] = form

        self.load()

    def load(self):
        """Load settings from the form.

        This will populate initial fields based on the site configuration.
        """
        super(AuthenticationSettingsForm, self).load()

        self.fields['auth_anonymous_access'].initial = \
            not self.siteconfig.get('auth_require_sitewide_login')
        self.fields['automatic_api_token_expiration'].initial = \
            self.siteconfig.get('client_token_expiration')

    def save(self):
        """Save the form.

        This will write the new configuration to the database. It will then
        force a site configuration reload.
        """
        self.siteconfig.set('auth_require_sitewide_login',
                            not self.cleaned_data['auth_anonymous_access'])
        self.siteconfig.set(
            'client_token_expiration',
            self.cleaned_data['automatic_api_token_expiration'])

        auth_backend = self.cleaned_data['auth_backend']

        if auth_backend in self.auth_backend_forms:
            self.auth_backend_forms[auth_backend].save()

        for form, enable_field_id in self._iter_sso_backend_forms():
            if self[enable_field_id].data:
                form.save()

        super(AuthenticationSettingsForm, self).save()

        # Reload any important changes into the Django settings.
        load_site_config()

    def is_valid(self):
        """Return whether the form is valid.

        This will check the validity of the fields on this form and on
        the selected authentication backend's settings form.

        Returns:
            bool:
            ``True`` if the main settings form and authentication backend's
            settings form is valid. ``False`` if either form is invalid.
        """
        if not super(AuthenticationSettingsForm, self).is_valid():
            return False

        backend_id = self.cleaned_data['auth_backend']
        backend_form = self.auth_backend_forms[backend_id]

        if not backend_form.is_valid():
            return False

        for form, enable_field_id in self._iter_sso_backend_forms():
            if (self.cleaned_data[enable_field_id] and
                not form.is_valid()):
                return False

        return True

    def full_clean(self):
        """Clean and validate all form fields.

        This will clean and validate both this form and the selected
        authentication backend's settings form (or all settings forms, if this
        form has not been POSTed to).

        Raises:
            django.core.exceptions.ValidationError:
                One or more fields failed validation.
        """
        super(AuthenticationSettingsForm, self).full_clean()

        if self.data:
            # Note that this isn't validated yet, but that's okay given our
            # usage. It's a bit of a hack though.
            auth_backend = (self['auth_backend'].data or
                            self.fields['auth_backend'].initial)

            if auth_backend in self.auth_backend_forms:
                self.auth_backend_forms[auth_backend].full_clean()

            for form, enable_field_id in self._iter_sso_backend_forms():
                if (self[enable_field_id].data or
                    self.fields[enable_field_id].initial):
                    form.full_clean()
        else:
            for form in self.auth_backend_forms.values():
                form.full_clean()

            for form in self.sso_backend_forms.values():
                form.full_clean()

    def _iter_sso_backend_forms(self):
        """Yield the SSO backend forms.

        Yields:
            tuple:
            A 2-tuple of the SSO backend form and the name of the form field
            to enable that backend.
        """
        for sso_backend_id, form in self.sso_backend_forms.items():
            enable_field_id = '%s_enabled' % sso_backend_id

            yield form, enable_field_id

    class Meta:
        title = _('Authentication Settings')
        save_blacklist = ('auth_anonymous_access',)

        subforms = (
            {
                'subforms_attr': 'auth_backend_forms',
                'controller_field': 'auth_backend',
            },
            {
                'subforms_attr': 'sso_backend_forms',
                'controller_field': None,
                'enable_checkbox': True,
            },
        )

        fieldsets = (
            {
                'classes': ('wide',),
                'fields': ['auth_anonymous_access',
                           'auth_backend'],
            },
            {
                'title': _('API Authentication Settings'),
                'description': _(
                    'Manage how applications and services (e.g. RBTools) '
                    'authenticate to the Review Board API.'
                ),
                'classes': ('wide',),
                'fields': (
                    'automatic_api_token_expiration',
                ),
            },
        )



from django.conf import settings
from django.conf.urls.static import static
from django.urls import include, path, re_path
from django.views.generic import TemplateView
from djblets.util.views import HealthCheckView, cached_javascript_catalog
from pipeline import views as pipeline_views

from reviewboard.accounts import views as accounts_views
from reviewboard.admin import views as admin_views
from reviewboard.attachments import views as attachments_views
from reviewboard.datagrids.urls import urlpatterns as datagrid_urlpatterns
from reviewboard.extensions.base import get_extension_manager
from reviewboard.hostingsvcs.urls import urlpatterns as hostingsvcs_urlpatterns
from reviewboard.reviews import views as reviews_views
from reviewboard.webapi.resources import resources


extension_manager = get_extension_manager()


handler404 = 'django.views.defaults.page_not_found'
handler500 = 'django.views.defaults.server_error'


# Useful collections of URL names that may be interesting to callers.
# This is especially useful for any apply_to lists in hooks.

#: A list of all diff viewer URL names.
diffviewer_url_names = [
    'view-diff',
    'view-interdiff',
    'view-diff-revision',
]


#: A list of all reviewable URL names.
#:
#: This contains diffs, file attachments, and legacy screenshot URLs.
#:
#: It does not contain the review request URLs. For that, see
#: :py:data:`all_review_request_url_names`.
reviewable_url_names = diffviewer_url_names + [
    'file-attachment',
    'screenshot',
]


#: The URL name for the review request page.
main_review_request_url_name = 'review-request-detail'


#: A list of review request and diff viewer URL names.
#:
#: This does not contain URL names for other reviewable content, like
#: file attachments. For that, see :py:data:`all_review_request_url_names`.
review_request_url_names = diffviewer_url_names + [
    main_review_request_url_name,
]


#: A list of all review request and reviewable content URL names.
#:
#: Version Added:
#:     8.0
all_review_request_url_names = reviewable_url_names + review_request_url_names


# URLs global to all modes
urlpatterns = [
    path('admin/extensions/',
         include('djblets.extensions.urls'),
         kwargs={
             'extension_manager': extension_manager,
         }),
    path('admin/', include('reviewboard.admin.urls')),
    path('jsi18n/',
         cached_javascript_catalog,
         kwargs={
             'packages': ('reviewboard', 'djblets'),
         },
         name='js-catalog'),
    path('read-only/',
         TemplateView.as_view(template_name='read_only.html'),
         name='read-only'),
    path('health/',
         HealthCheckView.as_view(),
         name='health-check'),
]


urlpatterns += extension_manager.get_url_patterns()


# Add static media if running in DEBUG mode on a non-production host.
if settings.DEBUG and not settings.PRODUCTION:
    urlpatterns += static(settings.STATIC_DIRECTORY,
                          view=pipeline_views.serve_static,
                          show_indexes=True)
    urlpatterns += static(settings.MEDIA_DIRECTORY,
                          document_root=settings.MEDIA_ROOT,
                          view=attachments_views.serve_safe,
                          show_indexes=True)

    urlpatterns += [
        path('js-tests/',
             TemplateView.as_view(template_name='js/tests.html'),
             name='js-tests'),
        path('js-tests/extensions/',
             TemplateView.as_view(template_name='js/extension_tests.html'),
             name='js-extensions-tests'),
        path('debug/icons/',
             TemplateView.as_view(template_name='icons.html'),
             name='debug-icons'),
    ]


localsite_urlpatterns = [
    path('', reviews_views.RootView.as_view(), name='root'),

    path('api/', include(resources.root.get_url_patterns())),
    path('r/', include('reviewboard.reviews.urls')),

    # Support
    path('support/',
         admin_views.support_redirect,
         name='support'),

    # Users
    re_path(r'^users/(?P<username>[\w.@+-]+)/', include([
        # User info box
        path('infobox/',
             accounts_views.UserInfoboxView.as_view(),
             name='user-infobox'),

        # User file attachments
        path(r'file-attachments/<uuid:file_attachment_uuid>/',
            attachments_views.user_file_attachment,
            name='user-file-attachment'),
    ])),

    # Search
    path('search/', include('reviewboard.search.urls')),
]


localsite_urlpatterns += datagrid_urlpatterns
localsite_urlpatterns += hostingsvcs_urlpatterns


# Main includes
urlpatterns += [
    path('account/', include('reviewboard.accounts.urls')),
    path('oauth2/', include('reviewboard.oauth.urls')),
    re_path(r'^s/(?P<local_site_name>[\w\.-]+)/',
            include(localsite_urlpatterns)),
]


urlpatterns += localsite_urlpatterns



"""A hook for registering SCMTools."""

from __future__ import annotations

from django.utils.translation import gettext as _
from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint

from reviewboard.scmtools import scmtools_registry


class SCMToolHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for registering an SCMTool."""

    def initialize(self, scmtool_cls):
        """Initialize the hook.

        This will register the SCMTool.

        Args:
            scmtool_cls (type):
                The SCMTool class to register. This must be a subclass of
                :py:class:`~reviewboard.scmtools.core.SCMTool`.

        Raises:
            ValueError:
                The SCMTool's :py:attr:`~reviewboard.scmtools.core.SCMTool.
                scmtool_id` attribute was not set.
        """
        scmtool_id = scmtool_cls.scmtool_id

        if scmtool_id is None:
            raise ValueError(_('%s.scmtool_id must be set.')
                             % scmtool_cls.__name__)

        self.scmtool_id = scmtool_id
        scmtools_registry.register(scmtool_cls)

    def shutdown(self):
        """Shut down the hook.

        This will unregister the SCMTool.
        """
        scmtools_registry.unregister_by_attr('scmtool_id', self.scmtool_id)



"""Unit tests for reviewboard.admin.forms.EMailSettingsForm."""

import kgb
from django.contrib import messages
from django.core import mail
from djblets.mail.message import EmailMessage
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.admin.forms.email_settings import EMailSettingsForm, logger
from reviewboard.testing.testcase import TestCase


class EMailSettingsFormTestCase(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.admin.forms.EMailSettingsForm."""

    def test_save(self):
        """Testing EMailSettingsForm.save"""
        siteconfig = SiteConfiguration.objects.get_current()

        # Set defaults, so we know our changes are going to apply.
        siteconfig.set('mail_enable_autogenerated_header', False)
        siteconfig.set('mail_send_new_user_mail', False)
        siteconfig.set('mail_send_password_changed_mail', False)
        siteconfig.set('mail_send_review_close_mail', False)
        siteconfig.set('mail_send_review_mail', False)
        siteconfig.set('mail_use_tls', False)
        siteconfig.set('mail_default_from', 'noreply@example.com')
        siteconfig.set('mail_from_spoofing',
                       EmailMessage.FROM_SPOOFING_NEVER)
        siteconfig.set('mail_host', 'localhost')
        siteconfig.set('mail_host_password', '')
        siteconfig.set('mail_host_user', '')
        siteconfig.set('mail_port', 25)
        siteconfig.save()

        form = EMailSettingsForm(
            siteconfig,
            data={
                'mail_default_from': 'no-reply@mail.example.com',
                'mail_enable_autogenerated_header': True,
                'mail_from_spoofing': EmailMessage.FROM_SPOOFING_SMART,
                'mail_host': 'mail.example.com',
                'mail_host_password': 's3cr3t',
                'mail_host_user': 'mail-user',
                'mail_port': 123,
                'mail_send_new_user_mail': True,
                'mail_send_password_changed_mail': True,
                'mail_send_review_close_mail': True,
                'mail_send_review_mail': True,
                'mail_use_tls': True,
            })

        self.assertTrue(form.is_valid())
        form.save()

        # send_test_mail was not set.
        self.assertEqual(len(mail.outbox), 0)

        # Explicitly re-fetch this (without updating the cached copy) to
        # check the result.
        siteconfig = SiteConfiguration.objects.get(pk=siteconfig.pk)
        self.assertTrue(siteconfig.get('mail_enable_autogenerated_header'))
        self.assertTrue(siteconfig.get('mail_send_new_user_mail'))
        self.assertTrue(siteconfig.get('mail_send_password_changed_mail'))
        self.assertTrue(siteconfig.get('mail_send_review_close_mail'))
        self.assertTrue(siteconfig.get('mail_send_review_mail'))
        self.assertTrue(siteconfig.get('mail_use_tls'))
        self.assertEqual(siteconfig.get('mail_default_from'),
                         'no-reply@mail.example.com')
        self.assertEqual(siteconfig.get('mail_from_spoofing'),
                         EmailMessage.FROM_SPOOFING_SMART)
        self.assertEqual(siteconfig.get('mail_host'), 'mail.example.com')
        self.assertEqual(siteconfig.get('mail_host_password'), 's3cr3t')
        self.assertEqual(siteconfig.get('mail_host_user'), 'mail-user')
        self.assertEqual(siteconfig.get('mail_port'), 123)

    def test_save_with_send_test_mail_with_request(self):
        """Testing EMailSettingsForm.save with send_test_mail=True and with
        HTTP request
        """
        user = self.create_user(username='admin',
                                email='admin@corp.example.com')

        siteconfig = SiteConfiguration.objects.get_current()
        form = EMailSettingsForm(
            siteconfig,
            request=self.create_http_request(user=user),
            data={
                'mail_default_from': 'noreply@corp.example.com',
                'send_test_mail': True,
            })

        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(len(mail.outbox), 1)

        message = mail.outbox[0]
        self.assertEqual(message.subject, 'Review Board e-mail settings test')
        self.assertEqual(message.body,
                         "This is a test of the e-mail settings for the "
                         "Review Board server at http://example.com. If you "
                         "got this, you're all set!")
        self.assertEqual(message.to, ['admin@corp.example.com'])
        self.assertEqual(message.from_email, 'noreply@corp.example.com')

        # Explicitly re-fetch this (without updating the cached copy) to
        # check the result.
        siteconfig = SiteConfiguration.objects.get(pk=siteconfig.pk)
        self.assertNotIn('send_test_mail', siteconfig.settings)

    def test_save_with_send_test_mail_without_request(self):
        """Testing EMailSettingsForm.save with send_test_mail=True and without
        HTTP request
        """
        siteconfig = SiteConfiguration.objects.get_current()
        form = EMailSettingsForm(
            siteconfig,
            data={
                'mail_default_from': 'noreply@corp.example.com',
                'send_test_mail': True,
            })

        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(len(mail.outbox), 1)

        message = mail.outbox[0]
        self.assertEqual(message.subject, 'Review Board e-mail settings test')
        self.assertEqual(message.body,
                         "This is a test of the e-mail settings for the "
                         "Review Board server at http://example.com. If you "
                         "got this, you're all set!")
        self.assertEqual(message.to, ['admin@example.com'])
        self.assertEqual(message.from_email, 'noreply@corp.example.com')

        # Explicitly re-fetch this (without updating the cached copy) to
        # check the result.
        siteconfig = SiteConfiguration.objects.get(pk=siteconfig.pk)
        self.assertNotIn('send_test_mail', siteconfig.settings)

    def test_save_with_send_test_mail_with_request_and_error(self):
        """Testing EMailSettingsForm.save with send_test_mail=True with
        HTTP request and error sending e-mail
        """
        self.spy_on(logger.exception)
        self.spy_on(EmailMessage.send,
                    owner=EmailMessage,
                    op=kgb.SpyOpRaise(Exception('Kaboom!')))

        user = self.create_user(username='admin',
                                email='admin@corp.example.com')
        request = self.create_http_request(user=user)

        siteconfig = SiteConfiguration.objects.get_current()
        form = EMailSettingsForm(
            siteconfig,
            request=request,
            data={
                'mail_default_from': 'noreply@corp.example.com',
                'send_test_mail': True,
            })

        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(len(mail.outbox), 0)
        self.assertSpyCalledWith(logger.exception,
                                 'Failed to send test e-mail to %s: %s',
                                 'admin@corp.example.com',
                                 'Kaboom!')

        msgs = list(messages.get_messages(request))
        self.assertEqual(len(msgs), 1)

        msg = msgs[0]
        self.assertEqual(msg.level, messages.ERROR)
        self.assertEqual(msg.message,
                         'Failed to send the test e-mail: "Kaboom!". Check '
                         'the server logs for additional details.')

    def test_save_with_send_test_mail_without_request_and_error(self):
        """Testing EMailSettingsForm.save with send_test_mail=True without
        HTTP request and error sending e-mail
        """
        self.spy_on(logger.exception)
        self.spy_on(EmailMessage.send,
                    owner=EmailMessage,
                    op=kgb.SpyOpRaise(Exception('Kaboom!')))

        siteconfig = SiteConfiguration.objects.get_current()
        form = EMailSettingsForm(
            siteconfig,
            data={
                'mail_default_from': 'noreply@corp.example.com',
                'send_test_mail': True,
            })

        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(len(mail.outbox), 0)
        self.assertSpyCalledWith(logger.exception,
                                 'Failed to send test e-mail to %s: %s',
                                 'admin@example.com',
                                 'Kaboom!')



"""File-based storage backend for SSL/TLS certificates.

Version Added:
    6.0
"""

from __future__ import annotations

import json
import logging
import os
import re
from typing import TYPE_CHECKING, cast
from uuid import uuid4

import idna
from django.core.cache import cache
from django.core.files.utils import validate_file_name
from django.utils.text import slugify
from django.utils.translation import gettext as _, gettext_lazy
from djblets.cache.backend import cache_memoize, make_cache_key
from djblets.util.filesystem import safe_join
from djblets.util.functional import iterable_len, lazy_re_compile

from reviewboard.certs.cert import (CertDataFormat,
                                    CertPurpose,
                                    Certificate,
                                    CertificateBundle,
                                    CertificateFingerprints)
from reviewboard.certs.errors import (CertificateNotFoundError,
                                      CertificateStorageError)
from reviewboard.certs.storage.base import (BaseCertificateStorageBackend,
                                            BaseStoredCertificate,
                                            BaseStoredCertificateBundle,
                                            BaseStoredCertificateFingerprints,
                                            StorageStats)
from reviewboard.certs.utils import normalize_cert_hostname
from reviewboard.site.models import AnyOrAllLocalSites, LocalSite

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping, Sequence
    from typing import Any


logger = logging.getLogger(__name__)


class FileStoredDataMixin:
    """Mixin for all file-based stored data classes.

    This contains additional information and helpers for representing errors
    and parsing storage IDs.

    Version Added:
        6.0
    """

    #: The directory name where this class's data will be stored.
    storage_dir: str

    #: The display name of this stored data object.
    #:
    #: This will be used in error messages.
    storage_name: str

    #: A regex used to parse storage IDs.
    storage_id_re: re.Pattern

    @classmethod
    def parse_storage_id(
        cls,
        storage_id: str,
    ) -> Mapping[str, Any]:
        """Return data extracted from a storage ID.

        Args:
            storage_id (str):
                The storage ID to parse.

        Returns:
            dict:
            A dictionary containing parsed data for this stored data class.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                The storage ID could not be parsed.
        """
        m = cls.storage_id_re.match(storage_id)

        if not m:
            # There was an error with the ID. Report it.
            error_id = str(uuid4())
            logger.error('[%s] Invalid SSL/TLS CA %s file-based storage '
                         'ID "%s"',
                         error_id, cls.storage_name, storage_id)

            raise CertificateStorageError(
                _('Internal error parsing a SSL/TLS %(type)s storage ID. '
                  'Administrators can find details in the Review Board '
                  'server logs (error ID %(error_id)s).')
                % {
                    'error_id': error_id,
                    'type': cls.storage_name,
                })

        local_site_name = m.group('local_site')
        local_site: (LocalSite | None) = None

        if local_site_name is not None:
            try:
                local_site = LocalSite.objects.get(name=local_site_name)
            except LocalSite.DoesNotExist:
                # This is not ideal, but it at least gives us a semi-stable
                # result and avoids crashing unnecessarily.
                local_site = LocalSite(name=local_site_name)

        result: dict[str, Any] = dict(m.groupdict(),
                                      local_site=local_site)

        # This is somewhat hacky, in that we're assuming knowledge of a
        # key that isn't common to all stored data objects. However, it's
        # pretty harmless, and simplifies the implementation considerably.
        if 'port' in result:
            result['port'] = int(result['port'])

        return result


class FileStoredCertificate(FileStoredDataMixin, BaseStoredCertificate):
    """File-based storage for a certificate and private key.

    This will store identifying information on the certificate (hostname and
    port), along with file paths where the certificate and private key PEM
    files can be found.

    Version Added:
        6.0
    """

    storage_dir = 'certs'
    storage_name = 'certificate'
    storage_id_re = lazy_re_compile(
        r'^(?:(?P<local_site>[^:]+):)?'
        r'(?P<hostname>[^:]+):'
        r'(?P<port>\d+):'
        r'(?P<purpose>%(purposes)s)$'
        % {
            'purposes': '|'.join(
                purpose.value
                for purpose in CertPurpose
            ),
        }
    )

    ######################
    # Instance variables #
    ######################

    #: The path to the certificate file.
    _cert_file_path: str

    #: The path to the key file, if found.
    _key_file_path: str | None

    #: The hostname used as part of a storage ID.
    #:
    #: This is used internally in order to handle wildcard certificates,
    #: where the certificate hostname will be the requested hostname
    #: while the storage hostname will contain the wildcard.
    #:
    #: Version Added:
    #:     8.0
    _storage_hostname: str | None

    @classmethod
    def parse_storage_id(
        cls,
        storage_id: str,
    ) -> Mapping[str, Any]:
        """Return data extracted from a storage ID.

        Args:
            storage_id (str):
                The storage ID to parse.

        Returns:
            dict:
            A dictionary containing parsed data for this stored data class.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                The storage ID could not be parsed.
        """
        data = super().parse_storage_id(storage_id)

        return {
            **data,
            'purpose': CertPurpose(data['purpose']),
        }

    def __init__(
        self,
        *,
        cert_file_path: str,
        key_file_path: (str | None) = None,
        hostname: (str | None) = None,
        storage_hostname: (str | None) = None,
        **kwargs,
    ) -> None:
        """Initialize the stored certificate information.

        Args:
            cert_file_path (str):
                The local path to the certificate PEM file.

            key_file_path (str, optional):
                The local path to the private key PEM file, if available.

            hostname (str, optional):
                The hostname serving the certificate.

                Either this or ``certificate`` must be provided.

            storage_hostname (str, optional):
                The hostname used as part of a storage ID.

                This is provided in order to handle wildcard certificates,
                where the certificate hostname will be the requested hostname
                while the storage hostname will contain the wildcard.

            **kwargs (dict, optional):
                Additional keyword arguments to pass to the parent
                constructor.

        Raises:
            ValueError:
                One or more arguments were not provided or contained invalid
                values.
        """
        self._cert_file_path = cert_file_path
        self._key_file_path = key_file_path

        if storage_hostname:
            storage_hostname = normalize_cert_hostname(storage_hostname)

        self._storage_hostname = storage_hostname

        super().__init__(hostname=hostname,
                         **kwargs)

    def build_storage_id(self) -> str:
        """Return a new storage ID for the stored data.

        File storage IDs are always based on the hostname, port, purpose, and
        any Local Site name.

        Version Added:
            8.0

        Returns:
            str:
            The new storage ID.
        """
        storage_hostname = self._storage_hostname or self.hostname
        storage_id = f'{storage_hostname}:{self.port}'

        if local_site := self.local_site:
            storage_id = f'{local_site.name}:{storage_id}'

        storage_id = f'{storage_id}:{self.purpose.value}'

        return storage_id

    def get_cert_file_path(
        self,
        *,
        data_format: CertDataFormat = CertDataFormat.PEM,
    ) -> str:
        """Return the filesystem path to a certificate.

        Args:
            data_format (reviewboard.certs.cert.CertDataFormat, optional):
                The requested certificate data file format.

        Returns:
            str:
            The resulting file path.
        """
        return self._cert_file_path

    def get_key_file_path(
        self,
        *,
        data_format: CertDataFormat = CertDataFormat.PEM,
    ) -> str | None:
        """Return the filesystem path to a certificate private key.

        Not all certificates will have an associated private key. If one is
        not available, this will return ``None``.

        Args:
            data_format (reviewboard.certs.cert.CertDataFormat, optional):
                The requested private key data file format.

        Returns:
            str:
            The resulting file path, or ``None`` if a private key is not
            available.
        """
        return self._key_file_path

    def load_certificate(self) -> Certificate:
        """Load and return a Certificate from storage.

        Returns:
            reviewboard.certs.cert.Certificate:
            The resulting loaded certificate.

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error loading the certificate from storage.

                Details are in the error message.
        """
        # Allow exceptions to bubble up.
        return Certificate.create_from_files(
            hostname=self.hostname,
            port=self.port,
            purpose=self.purpose,
            cert_path=self.get_cert_file_path(),
            key_path=self.get_key_file_path(),
        )

    def __repr__(self) -> str:
        """Return a string representation of the stored certificate.

        Returns:
            str:
            The string representation.
        """
        return (
            f'<FileStoredCertificate(storage_id={self.storage_id!r},'
            f' hostname={self.hostname!r},'
            f' port={self.port!r},'
            f' purpose={self.purpose},'
            f' cert_file_path={self._cert_file_path!r},'
            f' key_file_path={self._key_file_path!r})>'
        )


class FileStoredCertificateBundle(FileStoredDataMixin,
                                  BaseStoredCertificateBundle):
    """File-based storage for a CA bundle.

    This will store identifying information on the CA bundle (its name),
    along with a file path where the bundle can be found.

    Version Added:
        6.0
    """

    storage_dir = 'cabundles'
    storage_name = 'CA bundle'
    storage_id_re = lazy_re_compile(
        r'^(?:(?P<local_site>[^:]+):)?'
        r'(?P<name>[^:]+)$'
    )

    ######################
    # Instance variables #
    ######################

    #: The path to the CA bundle file.
    _bundle_file_path: str

    def __init__(
        self,
        *,
        bundle_file_path: str,
        name: (str | None) = None,
        bundle: (CertificateBundle | None) = None,
        local_site: (LocalSite | None) = None,
        storage_id: (str | None) = None,
        **kwargs,
    ) -> None:
        """Initialize the stored CA bundle information>

        Args:
            bundle_file_path (str):
                The local path to the CA bundle PEM file.

            name (str, optional):
                The associated name of the CA bundle.

                Either this or ``bundle`` must be provided.

            bundle (reviewboard.certs.cert.CertificateBundle, optional):
                The CA bundle in storage.

                Either this or ``name`` must be provided.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site associated with the certificate.

            storage_id (str, optional):
                The ID of the certificate in storage.

                If not provided, one will be computed.

            **kwargs (dict, optional):
                Additional keyword arguments to pass to the parent
                constructor.
        """
        self._bundle_file_path = bundle_file_path

        super().__init__(bundle=bundle,
                         local_site=local_site,
                         name=name,
                         storage_id=storage_id,
                         **kwargs)

    def build_storage_id(self) -> str:
        """Return a new storage ID for the stored CA bundle.

        File storage IDs are always based on the bundle name and Local Site
        name.

        Version Added:
            8.0

        Returns:
            str:
            The new storage ID.
        """
        storage_id = self.name

        if local_site := self.local_site:
            storage_id = f'{local_site.name}:{storage_id}'

        return storage_id

    def get_bundle_file_path(
        self,
        *,
        data_format: CertDataFormat = CertDataFormat.PEM,
    ) -> str:
        """Return the filesystem path to a certificate bundle.

        Args:
            data_format (reviewboard.certs.cert.CertDataFormat, optional):
                The requested certificate bundle data file format.

        Returns:
            str:
            The resulting file path.
        """
        return self._bundle_file_path

    def load_bundle(self) -> CertificateBundle:
        """Load and return the CA bundle data from storage.

        Returns:
            reviewboard.certs.cert.CertificateBundle:
            The resulting loaded CA bundle.

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error loading the CA bundle from storage.

                Details are in the error message.
        """
        # Allow exceptions to bubble up.
        return CertificateBundle.create_from_file(
            name=self.name,
            path=self._bundle_file_path)

    def __repr__(self) -> str:
        """Return a string representation of the stored certificate bundle.

        Returns:
            str:
            The string representation.
        """
        return (
            f'<FileStoredCertificateBundle'
            f'(storage_id={self.storage_id!r},'
            f' name={self.name!r},'
            f' bundle_file_path={self._bundle_file_path!r})>'
        )


class FileStoredCertificateFingerprints(FileStoredDataMixin,
                                        BaseStoredCertificateFingerprints):
    """File-based storage for certificate fingerprints.

    This will store identifying information on the fingerprints (hostname and
    port), along with a file path where the serialized fingerprints data can be
    found.

    Version Added:
        6.0
    """

    storage_dir = 'fingerprints'
    storage_name = 'fingerprints'
    storage_id_re = lazy_re_compile(
        r'^(?:(?P<local_site>[^:]+):)?'
        r'(?P<hostname>[^:]+):'
        r'(?P<port>\d+)$'
    )

    ######################
    # Instance variables #
    ######################

    #: The path to the fingerprints file.
    _fingerprints_file_path: str

    def __init__(
        self,
        *,
        fingerprints_file_path: str,
        **kwargs,
    ) -> None:
        """Initialize the stored certificate fingerprints information.

        Args:
            fingerprints_file_path (str):
                The local path to the fingerprints data file.

            **kwargs (dict, optional):
                Additional keyword arguments to pass to the parent
                constructor.
        """
        self._fingerprints_file_path = fingerprints_file_path

        super().__init__(**kwargs)

    def build_storage_id(self) -> str:
        """Return a new storage ID for the stored fingerprints.

        File storage IDs are always based on the hostname, port, and
        Local Site name.

        Version Added:
            8.0

        Returns:
            str:
            The new storage ID.
        """
        storage_id = f'{self.hostname}:{self.port}'

        if local_site := self.local_site:
            storage_id = f'{local_site.name}:{storage_id}'

        return storage_id

    def load_fingerprints(self) -> CertificateFingerprints:
        """Load and return the certificate fingerprints data from storage.

        Returns:
            reviewboard.certs.cert.CertificateFingerprints:
            The resulting loaded certificate fingerprints

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error loading the fingerprints from storage.

                Details are in the error message.
        """
        try:
            with open(self._fingerprints_file_path, 'r') as fp:
                return CertificateFingerprints.from_json(json.load(fp))
        except (IOError, ValueError) as e:
            error_id = str(uuid4())
            logger.error('[%s] Error loading SSL/TLS certificate '
                         'fingerprints file "%s": %s',
                         error_id, self._fingerprints_file_path, e)

            raise CertificateStorageError(
                _('Error loading SSL/TLS certificate fingerprints. '
                  'Administrators can find details in the Review Board '
                  'server logs (error ID %(error_id)s).')
                % {
                    'error_id': error_id,
                })

    def __repr__(self) -> str:
        """Return a string representation of the stored fingerprints.

        Returns:
            str:
            The string representation.
        """
        return (
            f'<FileStoredCertificateFingerprints'
            f'(storage_id={self.storage_id!r},'
            f' hostname={self.hostname!r},'
            f' port={self.port!r},'
            f' fingerprints_file_path={self._fingerprints_file_path!r})>'
        )


class FileCertificateStorageBackend(BaseCertificateStorageBackend[
    FileStoredCertificate,
    FileStoredCertificateBundle,
    FileStoredCertificateFingerprints,
]):
    """File-based storage for SSL/TLS certificates.

    This storage backend will store certificates, CA bundles, and fingerprints
    in the following patterned locations under the certificate store directory
    for this backend::

        cabundles/<slug>.pem
        certs/<purpose>/<hostname>__<port>.crt
        certs/<purpose>/<hostname>__<port>.key
        certs/<purpose>/__.<hostname>__<port>.crt
        certs/<purpose>/__.<hostname>__<port>.key
        fingerprints/<hostname>__<port>.json
        sites/<local_site_name>/cabundles/<slug>.pem
        sites/<local_site_name>/certs/<purpose>/<hostname>__<port>.crt
        sites/<local_site_name>/certs/<purpose>/<hostname>__<port>.key
        sites/<local_site_name>/certs/<purpose>/__.<hostname>__<port>.crt
        sites/<local_site_name>/certs/<purpose>/__.<hostname>__<port>.key
        sites/<local_site_name>/fingerprints/<hostname>__<port>.json

    Iterating through data involves scanning these directories for information,
    and possibly opening each file. This can involve a lot of IO for large
    numbers of certificates, but is fine in most simple deployments. Using
    CA bundles, rather than individual certificates, can help keep usage
    performant.

    For multi-server deployments, Power Pack's synchronized certificate storage
    is recommended (available in Power Pack 6 and up).

    Version Added:
        6.0
    """

    backend_id = 'file'
    name = gettext_lazy('File-based storage')

    _cabundle_re = lazy_re_compile(r'(?P<basename>.+)\.pem')

    _cert_re = lazy_re_compile(
        r'(?P<basename>(?P<hostname>(?:__)?[A-Za-z0-9.-]+)__'
        r'(?P<port>\d+))\.crt'
    )

    _fingerprints_json_re = lazy_re_compile(
        r'(?P<hostname>(?:__)?[A-Za-z0-9.-]+)__(?P<port>\d+)\.json'
    )

    def get_stats(
        self,
        *,
        local_site: AnyOrAllLocalSites = None,
        **kwargs,
    ) -> StorageStats:
        """Return statistics on the certificates managed by the backend.

        This will include the total number of stored certificates, CA bundles,
        and verified fingerprints across zero, one, or all Local Sites.

        It also includes state UUIDs that represent the current states of the
        backend.

        Args:
            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                An optional LocalSite bound to the stats.

                If ``None``, the global site will be used.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            reviewboard.certs.storage.base.StorageStats:
            The computed or cached storage statistics.
        """
        def _gen_stats() -> StorageStats:
            cert_counts_by_purpose: dict[CertPurpose, int] = {}
            cert_count: int = 0

            for purpose in CertPurpose:
                count = iterable_len(self.iter_stored_certificates(
                    purpose=purpose,
                    local_site=local_site,
                ))
                cert_count += count
                cert_counts_by_purpose[purpose] = count

            return StorageStats(
                ca_bundle_count=iterable_len(
                    self.iter_stored_ca_bundles(
                        local_site=local_site)),
                cert_count=cert_count,
                cert_counts_by_purpose=cert_counts_by_purpose,
                fingerprint_count=iterable_len(
                    self.iter_stored_fingerprints(
                        local_site=local_site)),
                state_uuid=str(uuid4()))

        cache_key = self._build_stats_cache_key(local_site=local_site)

        return cast(StorageStats, cache_memoize(cache_key, _gen_stats))

    def add_ca_bundle(
        self,
        bundle: CertificateBundle,
        *,
        local_site: (LocalSite | None) = None,
        **kwargs,
    ) -> FileStoredCertificateBundle:
        """Add a root CA bundle to storage.

        The certificate bundle's name along with ``local_site`` are considered
        unique within the storage backend. If there's an existing bundle with
        this information, it will be replaced.

        Args:
            bundle (reviewboard.certs.cert.CertificateBundle):
                The bundle to add.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            FileStoredCertificateBundle:
            The resulting stored certificate bundle.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There as an error storing this certificate.
        """
        name = bundle.name

        bundle_file_path = self._build_ca_bundle_file_path(
            name=name,
            local_site=local_site,
            create_parents_if_missing=True)
        assert bundle_file_path

        # Allow exceptions to bubble up.
        bundle.write_bundle_file(bundle_file_path)

        self._invalidate_stats_cache(local_site=local_site)

        return FileStoredCertificateBundle(
            bundle=bundle,
            bundle_file_path=bundle_file_path,
            local_site=local_site,
            storage=self)

    def delete_ca_bundle(
        self,
        *,
        name: str,
        local_site: (LocalSite | None) = None,
        **kwargs,
    ) -> None:
        """Delete a root CA bundle from storage.

        Args:
            bundle (reviewboard.certs.cert.CertificateBundle):
                The bundle to add.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting this CA bundle.
        """
        bundle_file_path = self._build_ca_bundle_file_path(
            name=name,
            local_site=local_site,
            if_exists=True)

        if not bundle_file_path:
            raise CertificateNotFoundError()

        try:
            os.unlink(bundle_file_path)
        except IOError as e:
            error_id = str(uuid4())
            logger.error('[%s] Unable to delete SSL/TLS bundle file at '
                         '"%s": %s',
                         error_id, bundle_file_path, e)

            raise CertificateStorageError(
                _('Error deleting SSL/TLS CA bundle. Administrators can '
                  'find details in the Review Board server logs (error '
                  'ID %(error_id)s).')
                % {
                    'error_id': error_id,
                })

        self._invalidate_stats_cache(local_site=local_site)

    def delete_ca_bundle_by_id(
        self,
        storage_id: str,
        **kwargs,
    ) -> None:
        """Delete a root CA bundle from storage identified by ID.

        Args:
            storage_id (str):
                The ID of the CA bundle to delete.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting this CA bundle.
        """
        self.delete_ca_bundle(
            **FileStoredCertificateBundle.parse_storage_id(storage_id))

    def get_stored_ca_bundle(
        self,
        *,
        name: str,
        local_site: (LocalSite | None) = None,
        **kwargs,
    ) -> FileStoredCertificateBundle | None:
        """Return a root CA bundle in storage.

        Args:
            name (str):
                The unique name of the CA bundle in storage.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            FileStoredCertificateBundle:
            The stored certificate bundle in storage, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving this CA bundle.
        """
        bundle_file_path = self._build_ca_bundle_file_path(
            name=name,
            local_site=local_site,
            if_exists=True)

        if not bundle_file_path:
            return None

        return FileStoredCertificateBundle(
            name=name,
            bundle_file_path=bundle_file_path,
            local_site=local_site,
            storage=self)

    def get_stored_ca_bundle_by_id(
        self,
        storage_id: str,
        **kwargs,
    ) -> FileStoredCertificateBundle | None:
        """Return a root CA bundle in storage identified by ID.

        Args:
            storage_id (str):
                The ID of the CA bundle in storage.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            FileStoredCertificateBundle:
            The stored certificate bundle in storage, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving this CA bundle.
        """
        return self.get_stored_ca_bundle(
            **FileStoredCertificateBundle.parse_storage_id(storage_id))

    def iter_stored_ca_bundles(
        self,
        *,
        local_site: AnyOrAllLocalSites = None,
        start: int = 0,
        **kwargs,
    ) -> Iterator[FileStoredCertificateBundle]:
        """Iterate through all root CA bundles in storage.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            start (int, optional):
                The 0-based index within the list of root CA bundles to start
                iterating at.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Yields:
            FileStoredCertificateBundle:
            Each stored certificate bundle in storage.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error iterating through CA bundles.
        """
        entries = self._iter_stored_data_files(
            stored_data_cls=FileStoredCertificateBundle,
            file_pattern=self._cabundle_re,
            local_site=local_site,
            start=start)

        for entry_path, entry_local_site, m in entries:
            yield FileStoredCertificateBundle(
                storage=self,
                name=m.group('basename'),
                bundle_file_path=entry_path,
                local_site=entry_local_site)

    def get_ca_bundles_dir(
        self,
        *,
        local_site: (LocalSite | None) = None,
    ) -> str:
        """Return a path containing all CA bundle files.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the bundles.

                If ``None``, the global site will be used.

        Returns:
            str:
            The path to the bundles directory.
        """
        return self._build_data_dir_path(
            stored_data_cls=FileStoredCertificateBundle,
            local_site=local_site)

    def add_certificate(
        self,
        certificate: Certificate,
        *,
        local_site: (LocalSite | None) = None,
        **kwargs,
    ) -> FileStoredCertificate:
        """Add a certificate to storage.

        The certificate's hostname and port along with ``local_site`` are
        considered unique within the storage backend. If there's an existing
        certificate with this information, it will be removed.

        Args:
            certificate (reviewboard.certs.cert.Certificate):
                The certificate to add.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            FileStoredCertificate:
            The resulting stored certificate.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error storing this certificate.
        """
        hostname = certificate.hostname
        port = certificate.port
        purpose = certificate.purpose

        # Write the certificate.
        cert_file_path = self._build_cert_file_path(
            hostname=hostname,
            port=port,
            ext='crt',
            purpose=purpose,
            local_site=local_site,
            create_parents_if_missing=True)
        assert cert_file_path

        # Allow exceptions to bubble up.
        certificate.write_cert_file(cert_file_path)

        # Make sure any key is deleted.
        key_file_path = self._build_cert_file_path(
            hostname=hostname,
            port=port,
            ext='key',
            purpose=purpose,
            local_site=local_site)
        assert key_file_path

        if os.path.exists(key_file_path):
            try:
                os.unlink(key_file_path)
            except IOError as e:
                logger.error('Error deleting SSL/TLS certificate '
                             'private key file "%s": %s',
                             key_file_path, e)

        if certificate.key_data:
            # Write the private key. Allow exceptions to bubble up.
            certificate.write_key_file(key_file_path)
        else:
            key_file_path = None

        self._invalidate_stats_cache(local_site=local_site)

        # Build the storage ID for this certificate.
        return FileStoredCertificate(
            certificate=certificate,
            purpose=purpose,
            cert_file_path=cert_file_path,
            key_file_path=key_file_path,
            local_site=local_site,
            storage=self)

    def delete_certificate(
        self,
        *,
        hostname: str,
        port: int,
        purpose: CertPurpose,
        local_site: (LocalSite | None) = None,
        **kwargs,
    ) -> None:
        """Delete a certificate from storage.

        Version Changed:
            8.0:
            Added the ``purpose`` argument.

        Args:
            hostname (str):
                The hostname of the certificate in storage.

            port (int):
                The port on the host serving the certificate.

            purpose (reviewboard.certs.cert.CertPurpose):
                The purpose of the certificate to delete.

                This indicates whether the certificate to match is used
                to trust a remote server or authenticate Review Board.

                Version Added:
                    8.0

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting this certificate.
        """
        cert_file_path = self._build_cert_file_path(
            hostname=hostname,
            port=port,
            ext='crt',
            purpose=purpose,
            local_site=local_site,
            if_exists=True)

        if not cert_file_path:
            raise CertificateNotFoundError()

        key_file_path = self._build_cert_file_path(
            hostname=hostname,
            port=port,
            ext='key',
            purpose=purpose,
            local_site=local_site,
            if_exists=True)

        try:
            os.unlink(cert_file_path)
        except IOError as e:
            error_id = str(uuid4())
            logger.error('[%s] Unable to delete SSL/TLS certificate file at '
                         '"%s": %s',
                         error_id, cert_file_path, e)

            raise CertificateStorageError(
                _('Error deleting SSL/TLS certificate. Administrators can '
                  'find details in the Review Board server logs (error '
                  'ID %(error_id)s).')
                % {
                    'error_id': error_id,
                })

        if key_file_path:
            try:
                os.unlink(key_file_path)
            except IOError as e:
                error_id = str(uuid4())
                logger.error('[%s] Unable to delete SSL/TLS private key file '
                             'at "%s": %s',
                             error_id, cert_file_path, e)

                raise CertificateStorageError(
                    _('Error deleting SSL/TLS private key. Administrators '
                      'can find details in the Review Board server logs '
                      '(error ID %(error_id)s).')
                    % {
                        'error_id': error_id,
                    })

        self._invalidate_stats_cache(local_site=local_site)

    def delete_certificate_by_id(
        self,
        storage_id: str,
        **kwargs,
    ) -> None:
        """Delete a certificate from storage.

        Args:
            storage_id (str):
                The ID of the certificate to delete.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting this certificate.
        """
        self.delete_certificate(
            **FileStoredCertificate.parse_storage_id(storage_id))

    def get_stored_certificate(
        self,
        *,
        hostname: str,
        port: int,
        purpose: CertPurpose,
        local_site: (LocalSite | None) = None,
        **kwargs,
    ) -> FileStoredCertificate | None:
        """Return a certificate from storage.

        Version Changed:
            8.0:
            Added the ``purpose`` argument.

        Args:
            hostname (str):
                The hostname of the certificate in storage.

            port (int):
                The port on the host serving the certificate.

            purpose (reviewboard.certs.cert.CertPurpose):
                The purpose of the certificate to fetch.

                This indicates whether the certificate is used to trust a
                remote server or authenticate Review Board.

                Version Added:
                    8.0

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            FileStoredCertificate:
            The stored certificate in storage, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving this certificate.
        """
        cert_file_path = self._build_cert_file_path(
            hostname=hostname,
            port=port,
            ext='crt',
            purpose=purpose,
            local_site=local_site,
            if_exists=True)

        if cert_file_path:
            storage_hostname = hostname
        else:
            # Look for a wildcard cert.
            storage_hostname = self._build_wildcard_hostname(hostname)

            cert_file_path = self._build_cert_file_path(
                hostname=storage_hostname,
                port=port,
                ext='crt',
                purpose=purpose,
                local_site=local_site,
                if_exists=True)

            if not cert_file_path:
                return None

        key_file_path = self._build_cert_file_path(
            hostname=storage_hostname,
            port=port,
            ext='key',
            purpose=purpose,
            local_site=local_site,
            if_exists=True)

        return FileStoredCertificate(
            hostname=hostname,
            port=port,
            purpose=purpose,
            storage_hostname=storage_hostname,
            cert_file_path=cert_file_path,
            key_file_path=key_file_path,
            local_site=local_site,
            storage=self)

    def get_stored_certificate_by_id(
        self,
        storage_id: str,
        **kwargs,
    ) -> FileStoredCertificate | None:
        """Return a certificate from storage identified by ID.

        Args:
            storage_id (str):
                The ID of the certificate in storage.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            FileStoredCertificate:
            The stored certificate in storage, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving this certificate.
        """
        return self.get_stored_certificate(
            **FileStoredCertificate.parse_storage_id(storage_id))

    def iter_stored_certificates(
        self,
        *,
        purpose: CertPurpose,
        local_site: AnyOrAllLocalSites = None,
        start: int = 0,
        **kwargs,
    ) -> Iterator[FileStoredCertificate]:
        """Iterate through all certificates in storage.

        Version Changed:
            8.0:
            Added the ``purpose`` argument.

        Args:
            purpose (reviewboard.certs.cert.CertPurpose):
                The purpose of the certificates to filter for.

                Version Added:
                    8.0

            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                An optional LocalSite that owns the certificates.

                If ``None``, the global site will be used.

            start (int, optional):
                The 0-based index within the list of root CA bundles to start
                iterating at.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Yields:
            BaseStoredCertificate:
            Each stored certificate in storage.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error iterating through stored certificates.
        """
        key_file_path: str | None

        entries = self._iter_stored_data_files(
            stored_data_cls=FileStoredCertificate,
            file_pattern=self._cert_re,
            local_site=local_site,
            start=start,
            subdirs=(purpose.value,),
        )

        check_key = (purpose == CertPurpose.CLIENT)

        for cert_file_path, cert_local_site, m in entries:
            basename = m.group('basename')

            if check_key:
                key_file_path = os.path.join(os.path.dirname(cert_file_path),
                                             f'{basename}.key')

                if not os.path.exists(key_file_path):
                    key_file_path = None
            else:
                key_file_path = None

            hostname = self._decode_hostname_for_storage(
                m.group('hostname'),
                filename=cert_file_path,
            )

            if hostname:
                yield FileStoredCertificate(
                    hostname=hostname,
                    port=int(m.group('port')),
                    purpose=purpose,
                    cert_file_path=cert_file_path,
                    key_file_path=key_file_path,
                    local_site=cert_local_site,
                    storage=self,
                )

    def add_fingerprints(
        self,
        fingerprints: CertificateFingerprints,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
        **kwargs,
    ) -> FileStoredCertificateFingerprints:
        """Add verified certificate fingerprints to storage.

        The ``hostname``, ``port``, and ``local_site`` values taken together
        must be unique within a storage backend.

        Backends may store this together with the certificates or separately.

        Args:
            hostname (str):
                The hostname serving the certificate.

            port (int):
                The port on the host serving the certificate.

            fingerprints (reviewboard.certs.cert.CertificateFingerprints):
                The certificate fingerprints to add.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            BaseStoredCertificateFingerprints:
            The resulting stored certificate fingerprints.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error storing these fingerprints.
        """
        path = self._build_fingerprints_file_path(
            hostname=hostname,
            port=port,
            local_site=local_site,
            create_parents_if_missing=True)
        assert path

        # Read/write the fingerprints data.
        try:
            with open(path, 'w') as fp:
                json.dump(fingerprints.to_json(), fp)
        except IOError as e:
            error_id = str(uuid4())
            logger.error('[%s] Unable to write SSL/TLS fingerprints '
                         'file at "%s": %s',
                         error_id, path, e)

            raise CertificateStorageError(
                _('Error writing SSL/TLS certificate fingerprints. '
                  'Administrators can find details in the Review Board '
                  'server logs (error ID %(error_id)s).')
                % {
                    'error_id': error_id,
                })

        self._invalidate_stats_cache(local_site=local_site)

        return FileStoredCertificateFingerprints(
            fingerprints=fingerprints,
            hostname=hostname,
            port=port,
            fingerprints_file_path=path,
            local_site=local_site,
            storage=self)

    def delete_fingerprints(
        self,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
        **kwargs,
    ) -> None:
        """Delete certificate fingerprints from storage.

        Args:
            hostname (str):
                The hostname of the certificate in storage.

            port (int):
                The port on the host serving the certificate.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Raises:
            reviewboard.cert.errors.CertificateNotFoundError:
                The fingerprints were not found.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting these fingerprints.
        """
        fingerprints_file_path = self._build_fingerprints_file_path(
            hostname=hostname,
            port=port,
            local_site=local_site,
            if_exists=True)

        if not fingerprints_file_path:
            raise CertificateNotFoundError()

        try:
            os.unlink(fingerprints_file_path)
        except IOError as e:
            error_id = str(uuid4())
            logger.error('[%s] Unable to delete SSL/TLS certificate '
                         'fingerprints file at "%s": %s',
                         error_id, fingerprints_file_path, e)

            raise CertificateStorageError(
                _('Error deleting SSL/TLS certificate fingerprints. '
                  'Administrators can find details in the Review Board '
                  'server logs (error ID %(error_id)s).')
                % {
                    'error_id': error_id,
                })

        self._invalidate_stats_cache(local_site=local_site)

    def delete_fingerprints_by_id(
        self,
        storage_id: str,
        **kwargs,
    ) -> None:
        """Delete certificate fingerprints from storage identified by ID.

        Args:
            storage_id (str):
                The ID of the certificate fingerprints to delete.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting these fingerprints.
        """
        self.delete_fingerprints(
            **FileStoredCertificateFingerprints.parse_storage_id(storage_id))

    def get_stored_fingerprints(
        self,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
        **kwargs,
    ) -> FileStoredCertificateFingerprints | None:
        """Return certificate fingerprints from storage.

        Args:
            hostname (str):
                The hostname of the certificate in storage.

            port (int):
                The port on the host serving the certificate.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            BaseStoredCertificateFingerprints:
            The stored certificate fingerprints, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving these fingerprints.
        """
        path = self._build_fingerprints_file_path(
            hostname=hostname,
            port=port,
            local_site=local_site,
            if_exists=True)

        if not path:
            return None

        return FileStoredCertificateFingerprints(
            hostname=hostname,
            port=port,
            fingerprints_file_path=path,
            local_site=local_site,
            storage=self)

    def get_stored_fingerprints_by_id(
        self,
        storage_id: str,
        **kwargs,
    ) -> FileStoredCertificateFingerprints | None:
        """Return certificate fingerprints from storage identified by ID.

        Args:
            storage_id (str):
                The ID of the certificate fingerprints in storage.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            CertificateFingerprints:
            The stored certificate fingerprints, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving these fingerprints.
        """
        return self.get_stored_fingerprints(
            **FileStoredCertificateFingerprints.parse_storage_id(storage_id))

    def iter_stored_fingerprints(
        self,
        *,
        local_site: AnyOrAllLocalSites = None,
        start: int = 0,
        **kwargs,
    ) -> Iterator[FileStoredCertificateFingerprints]:
        """Iterate through all certificate fingerprints in storage.

        Args:
            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                An optional LocalSite that owns the certificate fingerprints.

                If ``None``, the global site will be used.

            start (int, optional):
                The 0-based index within the list of root CA bundles to start
                iterating at.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Yields:
            FileStoredCertificateFingerprints:
            Each stored certificate fingerprints in storage.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error iterating through these fingerprints.
        """

        entries = self._iter_stored_data_files(
            stored_data_cls=FileStoredCertificateFingerprints,
            file_pattern=self._fingerprints_json_re,
            local_site=local_site,
            start=start)

        for entry_path, entry_local_site, m in entries:
            hostname = self._decode_hostname_for_storage(
                m.group('hostname'),
                filename=entry_path,
            )

            if hostname:
                yield FileStoredCertificateFingerprints(
                    hostname=hostname,
                    port=int(m.group('port')),
                    fingerprints_file_path=entry_path,
                    local_site=entry_local_site,
                    storage=self,
                )

    def _invalidate_stats_cache(
        self,
        *,
        local_site: LocalSite | None,
    ) -> None:
        """Invalidate the file storage stats cache information.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site value associated with the stats key.
        """
        cache.delete_many([
            make_cache_key(self._build_stats_cache_key(
                local_site=local_site)),
            make_cache_key(self._build_stats_cache_key(
                local_site=LocalSite.ALL)),
        ])

    def _iter_stored_data_dirs(
        self,
        stored_data_cls: type[FileStoredDataMixin],
        *,
        local_site: AnyOrAllLocalSites = None,
    ) -> Iterator[tuple[str, LocalSite | None]]:
        """Iterate through data directories.

        This will yield absolute paths to directories where data may be
        stored, based on a data storage class.

        If a Local Site is provided, one path will be yielded containing the
        Local Site's directory within the stored data path.

        if no Local Site is provided, one path will be yielded containing the
        global site data directory within the stored data path.

        If all Local Sites are requested, then this will yield every Local
        Site path and the top-level path.

        Args:
            stored_data_cls (type):
                The stored data class representing files to iterate through.

            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                An optional Local Site value to filter by.

        Yields:
            tuple:
            A 2-tuple of results:

            Tuple:
                0 (str):
                    The path to the data directory.

                1 (reviewboard.site.models.LocalSite):
                    The path's associated Local Site, or ``None`` for the
                    global site.
        """
        if local_site is LocalSite.ALL:
            # Yield the top-level (non-Local Site) data.
            yield self._build_data_dir_path(stored_data_cls), None

            sites_path = os.path.join(self.storage_path, 'sites')

            if os.path.exists(sites_path):
                # There's Local Site data. Try to efficiently iterate through
                # any Local Sites for directory names we find and yield those
                # paths and Local Sites.
                with os.scandir(sites_path) as entries:
                    local_site_names = [
                        _path.name
                        for _path in entries
                        if _path.is_dir()
                    ]

                if local_site_names:
                    local_sites = (
                        LocalSite.objects
                        .filter(name__in=local_site_names)
                        .iterator()
                    )

                    for path_local_site in local_sites:
                        yield (
                            self._build_data_dir_path(
                                stored_data_cls,
                                local_site=path_local_site),
                            path_local_site,
                        )
        else:
            yield (
                self._build_data_dir_path(stored_data_cls,
                                          local_site=local_site),
                local_site,
            )

    def _iter_stored_data_files(
        self,
        stored_data_cls: type[FileStoredDataMixin],
        *,
        file_pattern: re.Pattern,
        local_site: AnyOrAllLocalSites = None,
        start: int = 0,
        subdirs: (Sequence[str] | None) = None,
    ) -> Iterator[tuple[str, LocalSite | None, re.Match]]:
        """Iterate through all files in the specified data directories.

        This will iterate through all the files in each data directory
        provided by :py:meth:`_iter_data_dirs`. Results are filtered by
        file pattern and by a starting index.

        Files not matching the pattern are ignored, and will not count toward
        identifying the starting index.

        Note that while a starting index may be supplied, it cannot
        immediately seek to the corresponding index in results. That means
        that this may be slower the further into the results the caller
        indexes. For large deployments where this may matter, Power Pack's
        synchronized certificate management is recommended.

        Version Changed:
            8.0:
            Added the ``subdirs`` argument.

        Args:
            stored_data_cls (type):
                The stored data class representing files to iterate through.

            file_pattern (re.Pattern):
                The regex file pattern to match.

                This may contain capture groups, which will be available in
                each yielded result.

            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                An optional Local Site value to filter by.

            start (int, optional):
                The starting index for file results.

            subdirs (list of str, optional):
                Subdirectories to include between the data file path and
                the filename.

                Version Added:
                    8.0

        Yields:
            tuple:
            A 3-tuple representing a matched file:

            Tuple:
                0 (str):
                    The absolute path to the file.

                1 (reviewboard.site.models.LocalSite):
                    The path's associated Local Site, or ``None`` for the
                    global site.

                2 (re.Match):
                    The regex match object for the supplied pattern.
        """
        i: int = 0
        dirs = self._iter_stored_data_dirs(stored_data_cls,
                                           local_site=local_site)

        for dir_entry, dir_local_site in dirs:
            if subdirs:
                dir_entry = os.path.join(dir_entry, *subdirs)

            if not os.path.exists(dir_entry):
                continue

            with os.scandir(dir_entry) as file_entries:
                norm_file_entries = sorted(
                    (
                        _file_entry
                        for _file_entry in file_entries
                        if _file_entry.is_file()
                    ),
                    key=lambda _file_entry: _file_entry.name)

                for file_entry in norm_file_entries:
                    m = file_pattern.match(file_entry.name)

                    if m and i >= start:
                        # This is a valid match. Yield to the caller.
                        yield file_entry.path, dir_local_site, m
                        i += 1

    def _build_data_dir_path(
        self,
        stored_data_cls: type[FileStoredDataMixin],
        *,
        local_site: (LocalSite | None) = None,
    ) -> str:
        """Return a path to a data file.

        Args:
            stored_data_cls (type):
                The stored data class representing files to iterate through.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional Local Site for the data directory.
        """
        path = self.storage_path

        if local_site:
            path = safe_join(path, 'sites', local_site.name)

        return os.path.join(path, stored_data_cls.storage_dir)

    def _build_data_file_path(
        self,
        *,
        stored_data_cls: type[FileStoredDataMixin],
        filename: str,
        local_site: (LocalSite | None) = None,
        create_parents_if_missing: bool = False,
        if_exists: bool = False,
        subdirs: (Sequence[str] | None) = None,
    ) -> str | None:
        """Return a path to a data file.

        Version Changed:
            8.0:
            Added the ``subdirs`` argument.

        Args:
            stored_data_cls (type):
                The stored data class representing files to iterate through.

            filename (str):
                The filename to append to the data directory.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional Local Site for the data directory.

            create_parents_if_missing (bool, optional):
                Whether to create all parent directories to the file, if
                it does not already exist.

            if_exists (bool, optional):
                Whether to only return a result if the path exists.

            subdirs (list of str, optional):
                Subdirectories to include between the data file path and
                the filename.

                Version Added:
                    8.0

        Returns:
            str:
            The resulting file path.

            This will be ``None`` if passing ``if_exists=True`` and the path
            does not exist.

        Raises:
            django.core.exceptions.SuspiciousFileOperation:
                One of the generated paths was outside of an expected
                directory.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error creating parent directories.
        """
        assert filename
        validate_file_name(filename)

        data_path = self._build_data_dir_path(stored_data_cls,
                                              local_site=local_site)

        if subdirs:
            data_path = os.path.join(data_path, *subdirs)

        file_path = safe_join(data_path, filename)

        if if_exists and not os.path.exists(file_path):
            return None

        if create_parents_if_missing and not os.path.exists(data_path):
            try:
                os.makedirs(data_path, 0o700, exist_ok=True)
            except Exception as e:
                error_id = str(uuid4())
                logger.error('[%s] Error creating SSL/TLS data directory '
                             '"%s": %s',
                             error_id, data_path, e)

                raise CertificateStorageError(
                    _('Error creating SSL/TLS data storage directory. '
                      'Administrators can find details in the Review Board '
                      'server logs (error ID %(error_id)s).')
                    % {
                        'error_id': error_id,
                    })

        return file_path

    def _build_cert_file_path(
        self,
        *,
        hostname: str,
        port: int,
        ext: str,
        purpose: CertPurpose,
        **kwargs,
    ) -> str | None:
        """Return a path to a certificate file.

        Version Changed:
            8.0:
            Added the ``purpose`` argument.

        Args:
            hostname (str):
                The hostname serving the certificate.

            port (int):
                The port on the host serving the certificate.

            ext (str):
                The extension for the filename.

            purpose (reviewboard.certs.cert.CertPurpose):
                The certificate purpose to include in the filename.

                Version Added:
                    8.0

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`_build_data_file_path`.

        Returns:
            str:
            The resulting certificate file path, or ``None`` if passing
            ``if_exists=True`` and the path does not exist.

        Raises:
            django.core.exceptions.SuspiciousFileOperation:
                One of the generated paths was outside of an expected
                directory.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error creating parent directories or encoding
                the hostname for storage.
        """
        return self._build_data_file_path(
            stored_data_cls=FileStoredCertificate,
            filename=self._build_host_filename(hostname=hostname,
                                               port=port,
                                               ext=ext),
            subdirs=(purpose.value,),
            **kwargs)

    def _build_ca_bundle_file_path(
        self,
        *,
        name: str,
        **kwargs,
    ) -> str | None:
        """Return a path to a CA bundle file.

        Args:
            name (str):
                The name of the bundle file.

                This must be in :term:`slug` format.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`_build_data_file_path`.

        Returns:
            str:
            The resulting CA bundle file path, or ``None`` if passing
            ``if_exists=True`` and the path does not exist.

        Raises:
            django.core.exceptions.SuspiciousFileOperation:
                One of the generated paths was outside of an expected
                directory.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error creating parent directories.
        """
        assert name == slugify(name)

        return self._build_data_file_path(
            stored_data_cls=FileStoredCertificateBundle,
            filename=f'{name}.pem',
            **kwargs)

    def _build_fingerprints_file_path(
        self,
        *,
        hostname: str,
        port: int,
        **kwargs,
    ) -> str | None:
        """Return a path to a certificate fingerprints file.

        Args:
            hostname (str):
                The hostname serving the certificate.

            port (int):
                The port on the host serving the certificate.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`_build_data_file_path`.

        Returns:
            str:
            The resulting certificate file path, or ``None`` if passing
            ``if_exists=True`` and the path does not exist.

        Raises:
            django.core.exceptions.SuspiciousFileOperation:
                One of the generated paths was outside of an expected
                directory.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error creating parent directories or encoding
                the hostname for storage.
        """
        return self._build_data_file_path(
            stored_data_cls=FileStoredCertificateFingerprints,
            filename=self._build_host_filename(hostname=hostname,
                                               port=port,
                                               ext='json'),
            **kwargs)

    def _build_host_filename(
        self,
        *,
        hostname: str,
        port: int,
        ext: str,
    ) -> str:
        """Return a base filename for a hostname, port, and extension.

        This will take care of normalizing wildcard hostnames to filenames.

        Args:
            hostname (str):
                The hostname for the certificate.

                This may be a wildcard hostname.

            port (int):
                The port for the certificate.

            ext (str):
                The file extension.

        Returns:
            str:
            The resulting filename.

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                A filename for this hostname could not be built.
        """
        hostname = self._encode_hostname_for_storage(hostname)

        return f'{hostname}__{port}.{ext}'

    def _build_wildcard_hostname(
        self,
        hostname: str,
    ) -> str:
        """Return a wildcard hostname for a given hostname.

        This will strip off the leading name on the hostname and replace it
        with a wildcard. A hostname of ``test.example.com`` will be converted
        to ``*.example.com``.

        Args:
            hostname (str):
                The hostname to turn into a wildcard.

        Returns:
            str:
            The resulting wildcard hostname.
        """
        hostname = hostname.split('.', 1)[1]

        return f'*.{hostname}'

    def _build_stats_cache_key(
        self,
        *,
        local_site: AnyOrAllLocalSites,
    ) -> str:
        """Return a cache key for file-based storage stats.

        This will not be prefixed or normalized. It's expected to be passed
        to :py:func:`~djblets.cache.backend.cache_memoize` or
        to :py:func:`~djblets.cache.backend.make_cache_key`.

        Args:
            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                An optional LocalSite bound to the stats.

                If ``None``, the global site will be used.

        Returns:
            str:
            The resulting cached key.
        """
        key = 'stats-file-cert-storage'

        if local_site is None:
            key = f'{key}'
        elif local_site is LocalSite.ALL:
            key = f'{key}:local-site-all'
        else:
            key = f'{key}:local-site-{local_site.pk}'

        return key

    def _encode_hostname_for_storage(
        self,
        hostname: str,
    ) -> str:
        """Return an encoded hostname for storage interaction.

        This encodes a hostname for usage in a filename for storage purposes.

        If there are issues encoding the hostname, an exception is raised
        for the caller to handle or propagate to the UI.

        Version Added:
            8.0

        Args:
            hostname (str):
                The hostname to encode.

        Returns:
            str:
            The encoded hostname.

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                The hostname could not be encoded for storage.
        """
        if hostname.startswith('*.'):
            prefix = '__.'
            hostname = hostname[2:]
        else:
            prefix = ''

        try:
            encoded_hostname = (
                idna.encode(hostname.rstrip('.'),
                            uts46=True)
                .decode('ascii')
            )
        except (UnicodeDecodeError, idna.IDNAError) as e:
            # Either encoding or decoding failed. Log and raise an exception.
            logger.error('Hostname %r cannot be encoded to a valid hostname '
                         'file for cert storage: %s',
                         hostname, e)

            raise CertificateStorageError(
                _(
                    'The hostname {hostname} contains invalid characters '
                    'and cannot be stored.'
                ).format(hostname=repr(hostname))
            ) from e

        return f'{prefix}{encoded_hostname}'

    def _decode_hostname_for_storage(
        self,
        encoded_hostname: str,
        *,
        filename: str,
    ) -> str | None:
        """Return a decoded hostname for storage interaction.

        This decodes an encoded hostname used in a filename back to a
        standard hostname.

        If there are any issues decoding the hostname, ``None`` will be
        returned. Unlike :py:meth:`_encode_hostname_for_storage`, this does
        not raise an exception, as unexpected state in the directory is just
        intended to be skipped and does not necessarily represent a fatal
        error.

        Version Added:
            8.0

        Args:
            encoded_hostname (str):
                The hostname to decode.

            filename (str):
                The name of the corresponding file, for error reporting
                purposes.

        Returns:
            str:
            The decoded hostname, or ``None`` if it could not be decoded.
        """
        if encoded_hostname.startswith('__.'):
            prefix = '*.'
            encoded_hostname = encoded_hostname[3:]
        else:
            prefix = ''

        try:
            hostname = idna.decode(encoded_hostname.encode('ascii'),
                                   uts46=True)
        except (UnicodeEncodeError, idna.IDNAError) as e:
            # Either encoding or decoding failed. Return a None result to
            # skip this.
            logger.warning('File %r cannot be decoded as a valid hostname '
                           'file for cert storage: %s',
                           filename, e)

            return None

        return f'{prefix}{hostname}'



"""Add FileDiff.parent_diff and rename FileDiff.diff_base64 to diff.

``FileDiff.diff_base64`` is also changed to use a
:py:class:`djblets.db.fields.Base64Field`.

Version Added:
    1.0
"""

from django_evolution.mutations import AddField, RenameField
from djblets.db.evolution import FakeChangeFieldType
from djblets.db.fields import Base64Field


MUTATIONS = [
    FakeChangeFieldType('FileDiff', 'diff_base64', Base64Field),
    RenameField('FileDiff', 'diff_base64', 'diff', db_column='diff_base64'),
    AddField('FileDiff', 'parent_diff', Base64Field, initial="",
             db_column='parent_diff_base64'),
]



from djblets.testing.decorators import add_fixtures
from djblets.webapi.errors import INVALID_FORM_DATA
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.notifications.models import WebHookTarget
from reviewboard.site.models import LocalSite
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (webhook_item_mimetype,
                                                webhook_list_mimetype)
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.mixins_extra_data import (ExtraDataItemMixin,
                                                        ExtraDataListMixin)
from reviewboard.webapi.tests.urls import (get_webhook_item_url,
                                           get_webhook_list_url)


def compare_item(self, item_rsp, webhook):
    self.assertEqual(item_rsp['id'], webhook.pk)
    self.assertEqual(item_rsp['enabled'], webhook.enabled)
    self.assertEqual(item_rsp['url'], webhook.url)
    self.assertEqual(item_rsp['custom_content'],
                     webhook.custom_content)
    self.assertEqual(item_rsp['secret'],
                     webhook.secret)
    self.assertEqual(
        resources.webhook.parse_apply_to_field(item_rsp['apply_to'], None),
        webhook.apply_to)

    self.assertEqual(
        set(item['title'] for item in item_rsp['repositories']),
        set(repo.name for repo in webhook.repositories.all()))

    self.assertEqual(item_rsp['events'], webhook.events)
    self.assertEqual(item_rsp['extra_data'], webhook.extra_data)


class ResourceListTests(ExtraDataListMixin, BaseWebAPITestCase,
                        metaclass=BasicTestsMetaclass):
    """Tests for the WebHookResource list resource."""

    resource = resources.webhook

    sample_api_url = 'webhooks/'

    basic_get_use_admin = True
    basic_post_use_admin = True

    fixtures = ['test_users']

    compare_item = compare_item

    def setup_http_not_allowed_item_test(self, user):
        return get_webhook_list_url()

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        webhook = self.create_webhook(with_local_site=with_local_site)

        if populate_items:
            items = [webhook]
        else:
            items = []

        return (get_webhook_list_url(local_site_name),
                webhook_list_mimetype,
                items)

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site, local_site_name,
                              post_valid_data):
        if post_valid_data:
            post_data = {
                'enabled': 0,
                'events': '*',
                'url': 'http://example.com',
                'encoding': 'application/json',
                'custom_content': '',
                'apply_to': 'all',
            }
        else:
            post_data = {}

        return (get_webhook_list_url(local_site_name),
                webhook_item_mimetype,
                post_data,
                [local_site_name])

    def check_post_result(self, user, rsp, local_site_name=None):
        self.assertIn('webhook', rsp)
        item_rsp = rsp['webhook']

        webhook = WebHookTarget.objects.get(pk=item_rsp['id'])

        if local_site_name is None:
            self.assertIsNone(webhook.local_site)
        else:
            self.assertEqual(webhook.local_site.name, local_site_name)

        self.compare_item(item_rsp, webhook)

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_post_with_repositories(self):
        """Testing the POST <url> API with custom repositories"""
        repositories = [
            self.create_repository(name='Repo 1'),
            self.create_repository(name='Repo 2'),
        ]

        self.user.is_superuser = True
        self.user.save()

        rsp = self.api_post(
            get_webhook_list_url(),
            {
                'enabled': 0,
                'events': '*',
                'url': 'http://example.com',
                'encoding': 'application/json',
                'apply_to': 'custom',
                'repositories': ','.join(
                    str(repo.pk)
                    for repo in repositories
                )
            },
            expected_mimetype=webhook_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        self.check_post_result(self.user, rsp)

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_post_with_local_site_and_repository_on_other_site(self):
        """Testing the POST <URL> API with a local site and custom
        repository in a different local site
        """
        local_site_1 = LocalSite.objects.create(name='local-site-1')
        local_site_2 = LocalSite.objects.create(name='local-site-2')

        for local_site in (local_site_1, local_site_2):
            local_site.admins.add(self.user)
            local_site.users.add(self.user)

        repositories = [
            self.create_repository(name='Repo 1', local_site=local_site_1),
            self.create_repository(name='Repo 2', local_site=local_site_2),
        ]

        rsp = self.api_post(
            get_webhook_list_url(local_site_1),
            {
                'enabled': 0,
                'events': '*',
                'url': 'http://example.com',
                'encoding': 'application/json',
                'custom_content': '',
                'apply_to': 'custom',
                'repositories': ','.join(
                    str(repo.pk)
                    for repo in repositories
                )
            },
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertTrue('err' in rsp)
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertEqual(rsp['err']['msg'], INVALID_FORM_DATA.msg)
        self.assertTrue('fields' in rsp)
        self.assertTrue('repositories' in rsp['fields'])
        self.assertEqual(rsp['fields']['repositories'],
                         ['Select a valid choice. 2 is not one of the '
                          'available choices.'])

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_post_with_local_site_and_repository_not_on_site(self):
        """Testing the POST <URL> API with a local site and custom
        repository not on a local site
        """
        local_site_1 = LocalSite.objects.create(name='local-site-1')
        local_site_2 = LocalSite.objects.create(name='local-site-2')

        for local_site in (local_site_1, local_site_2):
            local_site.admins.add(self.user)
            local_site.users.add(self.user)

        repositories = [
            self.create_repository(name='Repo 1', local_site=local_site_1),
            self.create_repository(name='Repo 2')
        ]

        rsp = self.api_post(
            get_webhook_list_url(local_site_1),
            {
                'enabled': 0,
                'events': '*',
                'url': 'http://example.com',
                'encoding': 'application/json',
                'custom_content': '',
                'apply_to': 'custom',
                'repositories': ','.join(
                    str(repo.pk)
                    for repo in repositories
                )
            },
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertTrue('err' in rsp)
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertEqual(rsp['err']['msg'], INVALID_FORM_DATA.msg)
        self.assertTrue('fields' in rsp)
        self.assertTrue('repositories' in rsp['fields'])
        self.assertEqual(rsp['fields']['repositories'],
                         ['Select a valid choice. 2 is not one of the '
                          'available choices.'])

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_post_repositories_local_site_but_webhook_not(self):
        """Testing the POST <URL> API without a local site for repositories
        that are in a local site
        """
        local_site = LocalSite.objects.create(name='local-site-1')

        self.user.is_superuser = True
        self.user.save()

        repository = self.create_repository(name='Repo 1',
                                            local_site=local_site)

        rsp = self.api_post(
            get_webhook_list_url(),
            {
                'enabled': 0,
                'events': '*',
                'url': 'http://example.com',
                'encoding': 'application/json',
                'custom_content': '',
                'apply_to': 'custom',
                'repositories': str(repository.pk),
            },
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertTrue('err' in rsp)
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertEqual(rsp['err']['msg'], INVALID_FORM_DATA.msg)
        self.assertTrue('fields' in rsp)
        self.assertTrue('repositories' in rsp['fields'])
        self.assertEqual(rsp['fields']['repositories'],
                         ['Select a valid choice. 1 is not one of the '
                          'available choices.'])

    @webapi_test_template
    def test_post_with_global_site_and_set_local_site(self):
        """Testing the POST <URL> API and attempting to set a LocalSite for a
        non-LocalSite WebHook is ignored
        """
        local_site = LocalSite.objects.create(name='local-site-1')

        self.user.is_superuser = True
        self.user.save()

        rsp = self.api_post(
            get_webhook_list_url(),
            {
                'enabled': 0,
                'events': '*',
                'url': 'http://example.com',
                'encoding': 'application/json',
                'custom_content': '',
                'apply_to': 'all',
                'local_site': local_site.pk,
            },
            expected_mimetype=webhook_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        webhook = WebHookTarget.objects.get(pk=rsp['webhook']['id'])
        self.assertIsNone(webhook.local_site)
        self.compare_item(rsp['webhook'], webhook)

    @webapi_test_template
    def test_post_with_local_site_and_set_local_site(self):
        """Testing the POST <URL> API and attempting to set a LocalSite for
        another LocalSite's WebHook is ignored
        """
        local_site_1 = LocalSite.objects.create(name='local-site-1')
        local_site_1.users.add(self.user)

        local_site_2 = LocalSite.objects.create(name='local-site-2')

        self.user.is_superuser = True
        self.user.save()

        rsp = self.api_post(
            get_webhook_list_url(local_site_name='local-site-1'),
            {
                'enabled': 0,
                'events': '*',
                'url': 'http://example.com',
                'encoding': 'application/json',
                'custom_content': '',
                'apply_to': 'all',
                'local_site': local_site_2.pk,
            },
            expected_mimetype=webhook_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        webhook = WebHookTarget.objects.get(pk=rsp['webhook']['id'])
        self.assertEqual(webhook.local_site_id, local_site_1.pk)
        self.compare_item(rsp['webhook'], webhook)

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_post_multiple_events(self):
        """Testing the POST <URL> API with multiple events"""
        self.user.is_superuser = True
        self.user.save()

        rsp = self.api_post(
            get_webhook_list_url(),
            {
                'enabled': 0,
                'events': 'review_request_closed,review_request_published',
                'url': 'http://example.com',
                'encoding': 'application/json',
                'custom_content': '',
                'apply_to': 'all'
            },
            expected_mimetype=webhook_item_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('webhook', rsp)
        self.compare_item(rsp['webhook'], WebHookTarget.objects.get())

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_post_no_events(self):
        """Testing the POST <URL> API with no events"""
        self.user.is_superuser = True
        self.user.save()

        rsp = self.api_post(
            get_webhook_list_url(),
            {
                'enabled': 0,
                'events': '',
                'url': 'http://example.com',
                'encoding': 'application/json',
                'custom_content': '',
                'apply_to': 'all'
            },
            expected_mimetype=webhook_item_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('webhook', rsp)
        self.compare_item(rsp['webhook'], WebHookTarget.objects.get())

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_post_all_events_and_more(self):
        """Testing the POST <URL> API with all events (*) and additional
        events
        """
        self.user.is_superuser = True
        self.user.save()

        rsp = self.api_post(
            get_webhook_list_url(),
            {
                'enabled': 0,
                'events': 'review_request_closed,*,review_request_published',
                'url': 'http://example.com',
                'encoding': 'application/json',
                'custom_content': '',
                'apply_to': 'all'
            },
            expected_mimetype=webhook_item_mimetype)

        webhook = WebHookTarget.objects.get()

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('webhook', rsp)
        self.compare_item(rsp['webhook'], webhook)

        self.assertListEqual(webhook.events, ['*'])

    @add_fixtures(['test_scmtools'])
    @webapi_test_template
    def test_post_empty_repositories(self):
        """Testing the POST <URL> API with an empty repositories field"""
        self.user.is_superuser = True
        self.user.save()

        rsp = self.api_post(
            get_webhook_list_url(),
            {
                'enabled': 0,
                'events': 'review_request_closed,*,review_request_published',
                'url': 'http://example.com',
                'encoding': 'application/json',
                'custom_content': '',
                'apply_to': 'custom',
                'repositories': '',
            },
            expected_mimetype=webhook_item_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('webhook', rsp)
        self.compare_item(rsp['webhook'], WebHookTarget.objects.get())


class ResourceItemTests(ExtraDataItemMixin, BaseWebAPITestCase,
                        metaclass=BasicTestsMetaclass):
    """Tests for the WebHookResource item resource."""

    resource = resources.webhook

    sample_api_url = 'webhooks/<id>/'

    basic_get_use_admin = True
    basic_delete_use_admin = True
    basic_put_use_admin = True

    fixtures = ['test_users']

    compare_item = compare_item

    def setup_http_not_allowed_list_test(self, user):
        webhook = self.create_webhook()

        return get_webhook_item_url(webhook.pk)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        webhook = self.create_webhook(with_local_site=with_local_site)

        return (get_webhook_item_url(webhook.pk, local_site_name),
                webhook_item_mimetype,
                webhook)

    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        webhook = self.create_webhook(with_local_site=with_local_site)

        return (get_webhook_item_url(webhook.pk, local_site_name),
                webhook_item_mimetype,
                {},
                webhook,
                [local_site_name])

    #
    # HTTP DELETE tests
    #

    def setup_basic_delete_test(self, user, with_local_site, local_site_name):
        webhook = self.create_webhook(with_local_site=with_local_site)

        return (get_webhook_item_url(webhook.pk, local_site_name),
                [webhook])

    def check_delete_result(self, user, webhook):
        self.assertRaises(WebHookTarget.DoesNotExist,
                          lambda: WebHookTarget.objects.get(pk=webhook.pk))

    #
    # HTTP PUT tests
    #

    def check_put_result(self, user, item_rsp, item, local_site_name=None):
        webhook = WebHookTarget.objects.get(pk=item_rsp['id'])

        if local_site_name is None:
            self.assertIsNone(webhook.local_site)
        else:
            self.assertEqual(webhook.local_site.name, local_site_name)

        self.compare_item(item_rsp, webhook)

    @webapi_test_template
    def test_put_with_global_site_and_set_local_site(self):
        """Testing the PUT <URL> API and attempting to set a LocalSite for a
        non-LocalSite WebHook is ignored
        """
        local_site = LocalSite.objects.create(name='local-site-1')

        self.user.is_superuser = True
        self.user.save()

        webhook = self.create_webhook()

        rsp = self.api_put(
            get_webhook_item_url(webhook.pk),
            {
                'local_site': local_site.pk,
            },
            expected_mimetype=webhook_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        webhook = WebHookTarget.objects.get(pk=rsp['webhook']['id'])
        self.assertIsNone(webhook.local_site)
        self.compare_item(rsp['webhook'], webhook)

    @webapi_test_template
    def test_put_with_local_site_and_set_local_site(self):
        """Testing the PUT <URL> API and attempting to set a LocalSite for a
        another LocalSite's WebHook is ignored
        """
        local_site_1 = LocalSite.objects.create(name='local-site-1')
        local_site_1.users.add(self.user)

        local_site_2 = LocalSite.objects.create(name='local-site-2')

        self.user.is_superuser = True
        self.user.save()

        webhook = self.create_webhook(local_site=local_site_1)

        rsp = self.api_put(
            get_webhook_item_url(webhook.pk, local_site_1.name),
            {
                'local_site': local_site_2.pk,
            },
            expected_mimetype=webhook_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        webhook = WebHookTarget.objects.get(pk=rsp['webhook']['id'])
        self.assertEqual(webhook.local_site_id, local_site_1.pk)
        self.compare_item(rsp['webhook'], webhook)



from django.contrib.auth.models import User
from django.core.files.uploadedfile import SimpleUploadedFile

from reviewboard.reviews.forms import (DefaultReviewerForm, GroupForm,
                                       UploadDiffForm)
from reviewboard.reviews.models import DefaultReviewer
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class DefaultReviewerFormTests(TestCase):
    """Unit tests for DefaultReviewerForm."""

    fixtures = ['test_scmtools']

    def setUp(self):
        super(DefaultReviewerFormTests, self).setUp()

        self.local_site = LocalSite.objects.create(name='test')

        # Create repositories for the test.
        self.local_site_repo = self.create_repository(
            name='Test 1',
            local_site=self.local_site)
        self.global_site_repo = self.create_repository(name='Test 2')

        # Create users for the test.
        self.local_site_user = User.objects.create_user(username='testuser1')
        self.local_site.users.add(self.local_site_user)

        self.global_site_user = User.objects.create_user(username='testuser2')

        # Create groups for the test.
        self.local_site_group = self.create_review_group(
            name='test1',
            local_site=self.local_site)
        self.global_site_group = self.create_review_group(name='test2')

    def test_without_localsite(self):
        """Testing DefaultReviewerForm without a LocalSite"""
        # Make sure the initial state and querysets are what we expect on init.
        form = DefaultReviewerForm()

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['repository'].queryset),
                         [self.local_site_repo, self.global_site_repo])
        self.assertEqual(list(form.fields['people'].queryset),
                         [self.local_site_user, self.global_site_user])
        self.assertEqual(list(form.fields['groups'].queryset),
                         [self.local_site_group, self.global_site_group])

        # Now test what happens when it's been fed data and validated.
        form = DefaultReviewerForm(data={
            'name': 'Test',
            'file_regex': '.*',
            'repository': [self.global_site_repo.pk],
            'people': [self.global_site_user.pk],
            'groups': [self.global_site_group.pk],
        })

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['repository'].queryset),
                         [self.local_site_repo, self.global_site_repo])
        self.assertEqual(list(form.fields['people'].queryset),
                         [self.local_site_user, self.global_site_user])
        self.assertEqual(list(form.fields['groups'].queryset),
                         [self.local_site_group, self.global_site_group])
        self.assertIsNone(form.fields['people'].widget.local_site_name)

        self.assertTrue(form.is_valid())

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['repository'].queryset),
                         [self.local_site_repo, self.global_site_repo])
        self.assertEqual(list(form.fields['people'].queryset),
                         [self.local_site_user, self.global_site_user])
        self.assertEqual(list(form.fields['groups'].queryset),
                         [self.local_site_group, self.global_site_group])

        default_reviewer = form.save()

        self.assertIsNone(default_reviewer.local_site)
        self.assertEqual(list(default_reviewer.repository.all()),
                         [self.global_site_repo])
        self.assertEqual(list(default_reviewer.people.all()),
                         [self.global_site_user])
        self.assertEqual(list(default_reviewer.groups.all()),
                         [self.global_site_group])

    def test_without_localsite_and_instance(self):
        """Testing DefaultReviewerForm without a LocalSite and editing instance
        """
        default_reviewer = DefaultReviewer.objects.create(
            name='Test',
            file_regex='.*',
            local_site=self.local_site)

        form = DefaultReviewerForm(
            data={
                'name': 'Test',
                'file_regex': '.*',
            },
            instance=default_reviewer)
        self.assertTrue(form.is_valid())

        new_default_reviewer = form.save()
        self.assertEqual(default_reviewer.pk, new_default_reviewer.pk)
        self.assertIsNone(new_default_reviewer.local_site)

    def test_without_localsite_and_with_local_site_user(self):
        """Testing DefaultReviewerForm without a LocalSite and User on a
        LocalSite
        """
        form = DefaultReviewerForm(data={
            'name': 'Test',
            'file_regex': '.*',
            'people': [self.local_site_user.pk],
        })

        # Note that unlike others, this scenario is allowed.
        self.assertTrue(form.is_valid())

    def test_without_localsite_and_with_local_site_group(self):
        """Testing DefaultReviewerForm without a LocalSite and Group on a
        LocalSite
        """
        form = DefaultReviewerForm(data={
            'name': 'Test',
            'file_regex': '.*',
            'groups': [self.local_site_group.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'groups': ['Select a valid choice. 1 is not one of the '
                           'available choices.'],
            })

    def test_without_localsite_and_with_local_site_repo(self):
        """Testing DefaultReviewerForm without a LocalSite and Repository on a
        LocalSite
        """
        form = DefaultReviewerForm(data={
            'name': 'Test',
            'file_regex': '.*',
            'repository': [self.local_site_repo.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'repository': ['A repository with ID 1 was not found.'],
            })

    def test_with_limited_localsite(self):
        """Testing DefaultReviewerForm limited to a LocalSite"""
        form = DefaultReviewerForm(limit_to_local_site=self.local_site)

        self.assertEqual(form.limited_to_local_site, self.local_site)
        self.assertNotIn('local_site', form.fields)
        self.assertEqual(list(form.fields['repository'].queryset),
                         [self.local_site_repo])
        self.assertEqual(list(form.fields['people'].queryset),
                         [self.local_site_user])
        self.assertEqual(list(form.fields['groups'].queryset),
                         [self.local_site_group])
        self.assertEqual(form.fields['people'].widget.local_site_name,
                         self.local_site.name)

    def test_with_limited_localsite_and_changing_site(self):
        """Testing DefaultReviewerForm limited to a LocalSite and changing
        LocalSite
        """
        site2 = LocalSite.objects.create(name='test-site-2')

        form = DefaultReviewerForm(
            data={
                'name': 'Test',
                'file_regex': '.*',
                'local_site': site2,
            },
            limit_to_local_site=self.local_site)

        self.assertEqual(form.limited_to_local_site, self.local_site)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['local_site'], self.local_site)

        default_reviewer = form.save()
        self.assertEqual(default_reviewer.local_site, self.local_site)

    def test_with_limited_localsite_and_compatible_instance(self):
        """Testing DefaultReviewerForm limited to a LocalSite and editing
        compatible instance
        """
        default_reviewer = DefaultReviewer.objects.create(
            name='Test',
            file_regex='.*',
            local_site=self.local_site)

        # This should just simply not raise an exception.
        DefaultReviewerForm(instance=default_reviewer,
                            limit_to_local_site=self.local_site)

    def test_with_limited_localsite_and_incompatible_instance(self):
        """Testing DefaultReviewerForm limited to a LocalSite and editing
        incompatible instance
        """
        default_reviewer = DefaultReviewer.objects.create(
            name='Test',
            file_regex='.*')

        error_message = (
            'The provided instance is not associated with a LocalSite '
            'compatible with this form. Please contact support.'
        )

        with self.assertRaisesMessage(ValueError, error_message):
            DefaultReviewerForm(instance=default_reviewer,
                                limit_to_local_site=self.local_site)

    def test_with_limited_localsite_and_invalid_user(self):
        """Testing DefaultReviewerForm limited to a LocalSite with a User
        not on the LocalSite
        """
        form = DefaultReviewerForm(
            data={
                'name': 'Test',
                'file_regex': '.*',
                'people': [self.global_site_user.pk],
            },
            limit_to_local_site=self.local_site)

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'people': ['A user with ID 2 was not found.'],
            })

    def test_with_limited_localsite_and_invalid_group(self):
        """Testing DefaultReviewerForm limited to a LocalSite with a Group
        not on the LocalSite
        """
        form = DefaultReviewerForm(
            data={
                'name': 'Test',
                'file_regex': '.*',
                'groups': [self.global_site_group.pk],
            },
            limit_to_local_site=self.local_site)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'groups': ['Select a valid choice. 2 is not one of the '
                           'available choices.'],
            })

    def test_with_limited_localsite_and_invalid_repo(self):
        """Testing DefaultReviewerForm limited to a LocalSite with a
        Repository not on the LocalSite
        """
        form = DefaultReviewerForm(
            data={
                'name': 'Test',
                'file_regex': '.*',
                'repository': [self.global_site_repo.pk],
            },
            limit_to_local_site=self.local_site)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'repository': ['A repository with ID 2 was not found.'],
            })

    def test_with_localsite_in_data(self):
        """Testing DefaultReviewerForm with a LocalSite in form data"""
        # Make sure the initial state and querysets are what we expect on init.
        form = DefaultReviewerForm()

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['repository'].queryset),
                         [self.local_site_repo, self.global_site_repo])
        self.assertEqual(list(form.fields['people'].queryset),
                         [self.local_site_user, self.global_site_user])
        self.assertEqual(list(form.fields['groups'].queryset),
                         [self.local_site_group, self.global_site_group])
        self.assertIsNone(form.fields['people'].widget.local_site_name)

        # Now test what happens when it's been fed data and validated.
        form = DefaultReviewerForm(data={
            'name': 'Test',
            'file_regex': '.*',
            'local_site': self.local_site.pk,
            'repository': [self.local_site_repo.pk],
            'people': [self.local_site_user.pk],
            'groups': [self.local_site_group.pk],
        })

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['repository'].queryset),
                         [self.local_site_repo, self.global_site_repo])
        self.assertEqual(list(form.fields['people'].queryset),
                         [self.local_site_user, self.global_site_user])
        self.assertEqual(list(form.fields['groups'].queryset),
                         [self.local_site_group, self.global_site_group])
        self.assertIsNone(form.fields['people'].widget.local_site_name)

        self.assertTrue(form.is_valid())

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['repository'].queryset),
                         [self.local_site_repo, self.global_site_repo])
        self.assertEqual(list(form.fields['people'].queryset),
                         [self.local_site_user, self.global_site_user])
        self.assertEqual(list(form.fields['groups'].queryset),
                         [self.local_site_group, self.global_site_group])
        self.assertIsNone(form.fields['people'].widget.local_site_name)

        default_reviewer = form.save()

        self.assertEqual(default_reviewer.local_site, self.local_site)
        self.assertEqual(list(default_reviewer.repository.all()),
                         [self.local_site_repo])
        self.assertEqual(list(default_reviewer.people.all()),
                         [self.local_site_user])
        self.assertEqual(list(default_reviewer.groups.all()),
                         [self.local_site_group])

    def test_with_localsite_in_data_and_instance(self):
        """Testing DefaultReviewerform with a LocalSite in form data and
        editing instance
        """
        default_reviewer = DefaultReviewer.objects.create(
            name='Test',
            file_regex='.*')

        form = DefaultReviewerForm(
            data={
                'name': 'Test',
                'file_regex': '.*',
                'local_site': self.local_site.pk,
            },
            instance=default_reviewer)
        self.assertTrue(form.is_valid())

        new_default_reviewer = form.save()
        self.assertEqual(default_reviewer.pk, new_default_reviewer.pk)
        self.assertEqual(new_default_reviewer.local_site, self.local_site)

    def test_with_localsite_in_data_and_invalid_user(self):
        """Testing DefaultReviewerForm with a LocalSite in form data and User
        not on the LocalSite
        """
        form = DefaultReviewerForm(data={
            'name': 'Test',
            'file_regex': '.*',
            'local_site': self.local_site.pk,
            'people': [self.global_site_user.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'people': ['A user with ID 2 was not found.'],
            })

    def test_with_localsite_in_data_and_invalid_group(self):
        """Testing DefaultReviewerForm with a LocalSite in form data and Group
        not on the LocalSite
        """
        form = DefaultReviewerForm(data={
            'name': 'Test',
            'file_regex': '.*',
            'local_site': self.local_site.pk,
            'groups': [self.global_site_group.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'groups': ['Select a valid choice. 2 is not one of the '
                           'available choices.'],
            })

    def test_with_localsite_in_data_and_invalid_repo(self):
        """Testing DefaultReviewerForm with a LocalSite in form data and
        Repository not on the LocalSite
        """
        form = DefaultReviewerForm(data={
            'name': 'Test',
            'file_regex': '.*',
            'local_site': self.local_site.pk,
            'repository': [self.global_site_repo.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'repository': [
                    'A repository with ID 2 was not found.',
                ],
            })

    def test_form_with_positional_argument(self):
        """Testing DefaultReviewerForm when passing data as a positional
        argument
        """
        # This was a regression caused by the change to add the new related
        # user selector.
        form = DefaultReviewerForm({
            'name': 'test',
            'file_regex': '.*',
        })
        self.assertTrue(form.is_valid())


class GroupFormTests(TestCase):
    """Unit tests for GroupForm."""

    def setUp(self):
        super(GroupFormTests, self).setUp()

        self.local_site = LocalSite.objects.create(name='test')

        # Create users for the test.
        self.local_site_user = User.objects.create_user(username='testuser1')
        self.local_site.users.add(self.local_site_user)

        self.global_site_user = User.objects.create_user(username='testuser2')

    def test_without_localsite(self):
        """Testing GroupForm without a LocalSite"""
        # Make sure the initial state and querysets are what we expect on init.
        form = GroupForm()

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.local_site_user, self.global_site_user])
        self.assertIsNone(form.fields['users'].widget.local_site_name)

        # Now test what happens when it's been fed data and validated.
        form = GroupForm(data={
            'name': 'test',
            'display_name': 'Test',
            'users': [self.global_site_user.pk],
        })

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.local_site_user, self.global_site_user])
        self.assertIsNone(form.fields['users'].widget.local_site_name)

        self.assertTrue(form.is_valid())

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.local_site_user, self.global_site_user])

        group = form.save()

        self.assertIsNone(group.local_site)
        self.assertEqual(list(group.users.all()), [self.global_site_user])

    def test_without_localsite_and_instance(self):
        """Testing GroupForm without a LocalSite and editing instance"""
        group = self.create_review_group(local_site=self.local_site)

        form = GroupForm(
            data={
                'name': 'test',
                'display_name': 'Test',
            },
            instance=group)
        self.assertTrue(form.is_valid())

        new_group = form.save()
        self.assertEqual(group.pk, new_group.pk)
        self.assertIsNone(new_group.local_site)

    def test_without_localsite_and_with_local_site_user(self):
        """Testing GroupForm without a LocalSite and User on a LocalSite"""
        form = GroupForm(data={
            'name': 'test',
            'display_name': 'Test',
            'users': [self.local_site_user.pk],
        })
        self.assertTrue(form.is_valid())

    def test_with_limited_localsite(self):
        """Testing GroupForm limited to a LocalSite"""
        form = GroupForm(limit_to_local_site=self.local_site)

        self.assertEqual(form.limited_to_local_site, self.local_site)
        self.assertNotIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.local_site_user])
        self.assertEqual(form.fields['users'].widget.local_site_name,
                         self.local_site.name)

    def test_with_limited_localsite_and_changing_site(self):
        """Testing GroupForm limited to a LocalSite and changing LocalSite"""
        site2 = LocalSite.objects.create(name='test-site-2')

        form = GroupForm(
            data={
                'name': 'test',
                'display_name': 'Test',
                'users': [self.local_site_user.pk],
                'local_site': site2.pk,
            },
            limit_to_local_site=self.local_site)

        self.assertEqual(form.limited_to_local_site, self.local_site)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['local_site'], self.local_site)

        group = form.save()
        self.assertEqual(group.local_site, self.local_site)

    def test_with_limited_localsite_and_compatible_instance(self):
        """Testing GroupForm limited to a LocalSite and editing compatible
        instance
        """
        group = self.create_review_group(local_site=self.local_site)

        # This should just simply not raise an exception.
        GroupForm(instance=group,
                  limit_to_local_site=self.local_site)

    def test_with_limited_localsite_and_incompatible_instance(self):
        """Testing GroupForm limited to a LocalSite and editing incompatible
        instance
        """
        group = self.create_review_group()

        error_message = (
            'The provided instance is not associated with a LocalSite '
            'compatible with this form. Please contact support.'
        )

        with self.assertRaisesMessage(ValueError, error_message):
            GroupForm(instance=group,
                      limit_to_local_site=self.local_site)

    def test_with_limited_localsite_and_invalid_user(self):
        """Testing GroupForm limited to a LocalSite with a User not on the
        LocalSite
        """
        form = GroupForm(
            data={
                'name': 'test',
                'display_name': 'Test',
                'users': [self.global_site_user.pk],
            },
            limit_to_local_site=self.local_site)

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'users': ['A user with ID 2 was not found.'],
            })

    def test_with_localsite_in_data(self):
        """Testing GroupForm with a LocalSite in form data"""
        # Make sure the initial state and querysets are what we expect on init.
        form = GroupForm()

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.local_site_user, self.global_site_user])
        self.assertIsNone(form.fields['users'].widget.local_site_name)

        # Now test what happens when it's been fed data and validated.
        form = GroupForm(data={
            'name': 'test',
            'display_name': 'Test',
            'local_site': self.local_site.pk,
            'users': [self.local_site_user.pk],
        })

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.local_site_user, self.global_site_user])
        self.assertIsNone(form.fields['users'].widget.local_site_name)

        self.assertTrue(form.is_valid())

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['users'].queryset),
                         [self.local_site_user, self.global_site_user])

        group = form.save()
        self.assertEqual(group.local_site, self.local_site)
        self.assertEqual(list(group.users.all()), [self.local_site_user])

    def test_with_localsite_in_data_and_instance(self):
        """Testing GroupForm with a LocalSite in form data and editing instance
        """
        group = self.create_review_group()

        form = GroupForm(
            data={
                'name': 'test',
                'display_name': 'Test',
                'local_site': self.local_site.pk,
            },
            instance=group)
        self.assertTrue(form.is_valid())

        new_group = form.save()
        self.assertEqual(group.pk, new_group.pk)
        self.assertEqual(new_group.local_site, self.local_site)

    def test_with_localsite_in_data_and_invalid_user(self):
        """Testing GroupForm with a LocalSite in form data and User not on the
        LocalSite
        """
        form = GroupForm(data={
            'name': 'test',
            'display_name': 'Test',
            'local_site': self.local_site.pk,
            'users': [self.global_site_user.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'users': ['A user with ID 2 was not found.'],
            })

    def test_form_with_positional_argument(self):
        """Testing GroupForm when passing data as a positional argument"""
        # This was a regression caused by the change to add the new related
        # user selector.
        form = GroupForm({
            'name': 'test',
            'display_name': 'Test',
        })

        self.assertTrue(form.is_valid())


class UploadDiffFormTests(TestCase):
    """Unit tests for UploadDiffForm."""

    fixtures = ['test_users', 'test_scmtools']

    def test_clean_with_no_history(self):
        """Testing UploadDiffForm.clean with a review request created without
        history support
        """
        review_request = self.create_review_request(create_repository=True)
        form = UploadDiffForm(
            review_request=review_request,
            data={
                'basedir': '',
            },
            files={
                'path': SimpleUploadedFile(
                    'diff', self.DEFAULT_GIT_FILEDIFF_DATA_DIFF),
            })

        self.assertTrue(form.is_valid())

    def test_clean_with_history(self):
        """Testing UploadDiffForm.clean with a review request created with
        history support
        """
        review_request = self.create_review_request(create_repository=True,
                                                    create_with_history=True)

        form = UploadDiffForm(
            review_request=review_request,
            data={
                'basedir': '',
            },
            files={
                'path': SimpleUploadedFile(
                    'diff', self.DEFAULT_GIT_FILEDIFF_DATA_DIFF),
            })

        self.assertFalse(form.is_valid())
        self.assertNotEqual(form.non_field_errors, [])



"""Evolution to add the scmtool_id field to Repository.

Version Added:
    5.0
"""

from django.db import models
from django_evolution.mutations import AddField


MUTATIONS = [
    AddField('Repository', 'scmtool_id', models.CharField, max_length=255,
             null=True),
]
